# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the base DBTask classes."""

import abc
from pathlib import Path
from typing import Any, ClassVar, cast, override
from unittest import mock

from django.test import override_settings

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts.models import TaskTypes, WorkRequestResults
from debusine.db.models import Collection, WorkRequest
from debusine.db.playground import scenarios
from debusine.db.tasks.base import DBTask, DBWorkerProxyTask
from debusine.db.tasks.models import DBInputCollectionSingle
from debusine.server.collections.debusine_task_configuration import Changes
from debusine.tasks import BaseExternalTask, TaskConfigError, inputs
from debusine.tasks.inputs.mock_resolver import MockFieldResolver
from debusine.tasks.models import BaseDynamicTaskData, BaseTaskData
from debusine.tasks.tests.helpers import SampleBaseExternalTask
from debusine.test.django import TestCase, preserve_db_task_registry
from debusine.test.test_utils import preserve_task_registry


class AbstractTaskWithInputs(SampleBaseExternalTask[BaseTaskData], abc.ABC):
    """Task class with an input field."""

    TASK_TYPE = TaskTypes.WORKER
    TASK_VERSION = 1

    test_field = inputs.DebusineFQDNInput()


class AbstractDBTaskWithInputs(DBTask[BaseTaskData], abc.ABC):
    """DBTask class with an input field."""

    TASK_TYPE = TaskTypes.WORKER
    TASK_VERSION = 1

    test_field = inputs.DebusineFQDNInput()

    @override
    def get_task_configuration_subject_context(
        self,
    ) -> tuple[str | None, str | None]:
        raise NotImplementedError()

    @override
    def _execute(self) -> WorkRequestResults:
        raise NotImplementedError()


class DBTaskTests(TestCase):
    """Tests for :py:class:`DBTask`."""

    scenario = scenarios.DefaultContext()

    @preserve_db_task_registry()
    def make_concrete_dbtask(self) -> DBTask[BaseTaskData]:
        """Instantiate a concrete instance of DBTask."""

        class ConcreteDBTask(DBTask[BaseTaskData, BaseDynamicTaskData]):
            """Concrete implementation of DBTask."""

            TASK_TYPE = TaskTypes.WORKER
            TASK_VERSION = 1

            @override
            def get_task_configuration_subject_context(
                self,
            ) -> tuple[str | None, str | None]:
                raise NotImplementedError()

            @override
            def _execute(self) -> WorkRequestResults:
                raise NotImplementedError()

        work_request = self.playground.create_worker_task(
            task_name="concretedbtask"
        )
        return work_request.get_task()

    @preserve_db_task_registry()
    def make_task_with_inputs(self) -> AbstractDBTaskWithInputs:
        class TaskWithInputs(AbstractDBTaskWithInputs):
            """Concrete version of AbstractDBTaskWithInputs."""

        work_request = self.playground.create_worker_task(
            task_name="taskwithinputs",
        )
        return cast(AbstractDBTaskWithInputs, work_request.get_task())

    def test_is_valid_task_name(self) -> None:
        self.assertTrue(DBTask.is_valid_task_name(TaskTypes.WORKER, "sbuild"))

    def test_is_valid_task_name_case_insensitive(self) -> None:
        self.assertTrue(DBTask.is_valid_task_name(TaskTypes.WORKER, "Sbuild"))

    def test_is_valid_task_name_invalid_type(self) -> None:
        self.assertFalse(
            DBTask.is_valid_task_name(
                cast(TaskTypes, "does-not-exist"), "sbuild"
            )
        )

    def test_is_valid_task_name_invalid(self) -> None:
        self.assertFalse(
            DBTask.is_valid_task_name(TaskTypes.WORKER, "not_a_real_task")
        )

    def test_task_names_worker(self) -> None:
        task_names = DBTask.task_names(TaskTypes.WORKER)
        self.assertGreater(len(task_names), 0)
        self.assertIn("noop", task_names)

    def test_worker_task_names(self) -> None:
        task_names = DBTask.worker_task_names()
        self.assertGreater(len(task_names), 0)
        self.assertIn("noop", task_names)

    def test_is_worker_task(self) -> None:
        self.assertTrue(DBTask.is_worker_task("noop"))

    def test_is_worker_task_false(self) -> None:
        self.assertFalse(DBTask.is_worker_task("nonexistent"))

    def test_class_for_name_invalid_type(self) -> None:
        with self.assertRaisesRegex(
            ValueError, "'not-existing-type' is not a registered task type"
        ):
            DBTask.class_from_name(
                cast(TaskTypes, "not-existing-type"), "non-existing-class"
            )

    def test_class_for_name_sbuild(self) -> None:
        self.assertEqual(
            DBTask.class_from_name(TaskTypes.WORKER, 'sBuIlD').name, "sbuild"
        )

    def test_class_for_name_no_class(self) -> None:
        with self.assertRaisesRegex(
            ValueError,
            "'non-existing-class' is not a registered Worker task_name",
        ):
            DBTask.class_from_name(TaskTypes.WORKER, 'non-existing-class')

    def test_get_label_no_dynamic_data(self) -> None:
        wr = self.playground.create_worker_task(task_name="noop")
        self.assertEqual(wr.get_task().get_label(), "noop")

    def test_get_label_no_parameter_summary(self) -> None:
        wr = self.playground.create_worker_task(
            task_name="noop", mark_running=True
        )
        self.assertEqual(wr.get_task().get_label(), "noop")

    def test_get_label_parameter_summary(self) -> None:
        wr = self.playground.create_worker_task(
            task_name="noop", mark_running=True
        )
        assert wr.dynamic_task_data is not None
        wr.dynamic_task_data["parameter_summary"] = "summary"
        wr.save()
        self.assertEqual(wr.get_task().get_label(), "noop: summary")

    def test_get_description(self) -> None:
        task = self.make_concrete_dbtask()
        self.assertEqual(task.get_description(), task.__doc__)

    def test_get_description_no_docstring(self) -> None:
        task = self.make_concrete_dbtask()
        task.__doc__ = None
        self.assertEqual(task.get_description(), task.get_label())

    def test_compute_scheduler_tags(self) -> None:
        self.playground.create_group("test", [self.scenario.user])
        self.playground.create_group(
            "test", [self.scenario.user], workspace=self.scenario.workspace
        )
        task = self.make_concrete_dbtask()
        ws = self.scenario.workspace
        provided, required = task.compute_scheduler_tags()
        assert provided.sets == {
            ttags.ProvenanceProvided.SYSTEM: {
                ttags.SCOPE_PREFIX + ws.scope.name,
                ttags.WORKSPACE_PREFIX + f"{ws.scope.name}:{ws.name}",
                ttags.GROUP_PREFIX + f"{ws.scope.name}::test",
                ttags.GROUP_PREFIX + f"{ws.scope.name}:{ws.name}:test",
            },
        }
        assert required.sets == {
            wtags.ProvenanceRequired.SYSTEM: {
                wtags.WORKER_TYPE_EXTERNAL,
                wtags.TASK_PREFIX
                + f"{task.TASK_TYPE.lower()}:concretedbtask:version:1",
            },
        }

    def test_get_input_artifacts_ids_uses_dynamic_data(self) -> None:
        task = self.make_concrete_dbtask()
        task.dynamic_data = mock.Mock()
        task.dynamic_data.get_input_artifacts_ids.return_value = [1, 2]

        self.assertEqual(task.get_input_artifacts_ids(), [1, 2])

    def test_get_input_artifacts_ids_from_wrapped_task(self) -> None:
        work_request = self.playground.create_worker_task(task_name="noop")
        task = work_request.get_task()
        self.assertIsInstance(task, DBWorkerProxyTask)

        with mock.patch(
            "debusine.tasks.BaseExternalTask.get_input_artifacts_ids",
            return_value=[3, 4],
        ):
            self.assertEqual(task.get_input_artifacts_ids(), [3, 4])

    def test_compute_dynamic_data_saves_fields(self) -> None:
        task = self.make_task_with_inputs()
        resolver = MockFieldResolver(
            test_field="debusine.example.org", task_configuration=None
        )
        resolver.resolve_inputs(task)
        ddata = task.compute_dynamic_data()
        self.assertEqual(
            ddata.task_input_fields,
            {"test_field": "debusine.example.org", "task_configuration": None},
        )


class DBProxyTaskTests(TestCase):
    """Tests for :py:class:`DBProxyTask`."""

    scenario = scenarios.DefaultContext()

    @preserve_task_registry()
    @preserve_db_task_registry()
    def make_task_with_inputs(
        self, dynamic_task_data: dict[str, Any] | None = None
    ) -> DBTask[BaseTaskData]:
        class TaskWithInputs(AbstractTaskWithInputs):
            """Concrete version of AbstractTaskWithInputs."""

        DBWorkerProxyTask.populate_task_registry()

        work_request = self.playground.create_worker_task(
            task_name="taskwithinputs",
        )
        if dynamic_task_data is not None:
            work_request.dynamic_task_data = dynamic_task_data
            work_request.save()
        return work_request.get_task()

    @preserve_task_registry()
    @preserve_db_task_registry()
    def test_populate_task_registry(self) -> None:
        # samplebaseexternaltask does not exist
        self.assertNotIn(
            "sampleexternaltask",
            BaseExternalTask._sub_tasks[TaskTypes.WORKER],
        )
        self.assertNotIn(
            "sampleexternaltask", DBTask._sub_tasks[TaskTypes.WORKER]
        )

        class SampleExternalTask(
            SampleBaseExternalTask[BaseTaskData],
        ):
            """Sample class to test BaseExternalTask class."""

            @override
            def run(self, execute_directory: Path) -> WorkRequestResults:
                """Unused abstract method from BaseExternalTask."""
                raise NotImplementedError()

        # samplebaseexternaltask is registered with BaseExternalTask only
        self.assertIn(
            "sampleexternaltask",
            BaseExternalTask._sub_tasks[TaskTypes.WORKER],
        )
        self.assertNotIn(
            "sampleexternaltask", DBTask._sub_tasks[TaskTypes.WORKER]
        )

        DBWorkerProxyTask.populate_task_registry()

        # samplebaseexternaltask is registered with BaseExternalTask and DBTask
        self.assertIn(
            "sampleexternaltask",
            BaseExternalTask._sub_tasks[TaskTypes.WORKER],
        )
        self.assertIn("sampleexternaltask", DBTask._sub_tasks[TaskTypes.WORKER])

    def test_get_provided_worker_tags(self) -> None:
        external_worker_get_tags = self.enterContext(
            self.spy("debusine.tasks.sbuild.Sbuild.get_provided_worker_tags")
        )
        self.mock_is_command_available({"sbuild": False})
        sbuild_task_cls = DBTask.class_from_name(TaskTypes.WORKER, "sbuild")
        self.assertEqual(
            sbuild_task_cls.get_provided_worker_tags(),
            {
                wtags.TASK_PREFIX
                + f"{sbuild_task_cls.TASK_TYPE.lower()}:{sbuild_task_cls.name}"
                f":version:{sbuild_task_cls.TASK_VERSION}",
            },
        )
        external_worker_get_tags.assert_called_once_with()

    def test_get_input_artifacts_ids_from_fields(self) -> None:
        work_request = self.playground.create_worker_task(task_name="noop")
        task = work_request.get_task()
        with (
            mock.patch(
                "debusine.tasks.BaseExternalTask.get_input_artifacts_ids",
                return_value=[1, 2, 3],
            ),
        ):
            self.assertEqual(
                task.get_input_artifacts_ids_from_fields(), [1, 2, 3]
            )

    @override_settings(DEBUSINE_FQDN="debusine.example.org")
    def test_compute_dynamic_data_saves_fields(self) -> None:
        # TODO: update test when DBTask also has a field
        task = self.make_task_with_inputs()
        task.resolve_inputs(inputs.Stage.RUNNING)
        ddata = task.compute_dynamic_data()
        assert ddata.task_input_fields == {
            "test_field": "debusine.example.org",
            "task_configuration": DBInputCollectionSingle.from_collection(
                self.scenario.default_task_configuration_collection,
                lookup="default@debusine:task-configuration",
                label="task_configuration",
            ).model_dump(),
        }

    def test_input_field_values_from_task_data(self) -> None:
        # TODO: update test when DBTask also has a field
        test_value = "test_restore_inputs.example.org"
        with self.spy(
            "debusine.tasks.inputs.DebusineFQDNInput.deserialize_value"
        ) as deserialize_method:
            task = self.make_task_with_inputs(
                dynamic_task_data={
                    "task_input_fields": {"test_field": test_value}
                }
            )
        assert hasattr(task, "task")
        assert isinstance(task.task, AbstractTaskWithInputs)
        self.assertEqual(task.task.test_field, test_value)
        assert deserialize_method.call_args_list == [
            mock.call(task.task.inputs["test_field"], test_value)
        ]

    def test_input_field_values_from_task_data_ignore_spurious(self) -> None:
        # TODO: update test when DBTask also has a field
        test_value = "test_restore_inputs.example.org"
        task = self.make_task_with_inputs(
            dynamic_task_data={
                "task_input_fields": {
                    "test_field": test_value,
                    "other_field": "invalid",
                }
            }
        )
        assert hasattr(task, "task")
        assert isinstance(task.task, AbstractTaskWithInputs)
        self.assertEqual(task.task.test_field, test_value)
        self.assertFalse(hasattr(task, "other_field"))
        self.assertFalse(hasattr(task.task, "other_field"))


class ApplyTaskConfigurationTests(TestCase):
    """test :py:meth:`DBTask.apply_task_configuration`."""

    scenario = scenarios.DefaultContext()

    work_request: ClassVar[WorkRequest]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        environment_item = cls.playground.create_debian_environment()
        assert environment_item.artifact is not None
        environment = environment_item.artifact
        source = cls.playground.create_source_artifact()
        cls.work_request = cls.playground.create_sbuild_work_request(
            source=source, environment=environment, architecture="amd64"
        )
        cls.work_request.save()

    @override
    def setUp(self) -> None:
        super().setUp()
        self.apply_configuration = self.enterContext(
            mock.patch(
                "debusine.server.collections.debusine_task_configuration"
                ".apply_configuration"
            )
        )
        self.sample_task_data = self.work_request.task_data.copy()
        self.sample_dynamic_data = {
            "input_source_artifact_id": 1,
            "input_extra_binary_artifacts_ids": [],
            "binnmu_maintainer": None,
        }
        self.provide_tags = {"provided"}
        self.require_tags = {"required"}
        self.provide_tag_candidates = ttags.TaskProvidedTags()
        self.provide_tag_candidates.add(
            ttags.ProvenanceProvided.SYSTEM, {"provided"}
        )
        self.require_tag_candidates = wtags.TaskRequiredTags()
        self.require_tag_candidates.add(
            wtags.ProvenanceRequired.SYSTEM, {"required"}
        )

    def patch_task[T: DBTask[Any, Any]](self, task: T) -> T:
        """Mock relevant methods called by apply_task_configuration."""
        self.get_task_configuration_subject_context = self.enterContext(
            mock.patch.object(
                task,
                "get_task_configuration_subject_context",
                return_value=("subject", "context"),
            )
        )
        self.compute_scheduler_tags = self.enterContext(
            mock.patch.object(
                task,
                "compute_scheduler_tags",
                return_value=(
                    self.provide_tag_candidates,
                    self.require_tag_candidates,
                ),
            )
        )
        return task

    def get_task(
        self, task_configuration: Collection | None = None
    ) -> DBTask[Any, Any]:
        task = self.patch_task(self.work_request.get_task())
        # Fill the task_configuration field, which is a prerequisite before
        # apply_task_configuration can be called
        if task_configuration is None:
            task.__dict__["task_configuration"] = None
        else:
            task.__dict__["task_configuration"] = (
                DBInputCollectionSingle.from_collection(
                    task_configuration,
                    lookup=BaseTaskData.DEFAULT_TASK_CONFIGURATION_LOOKUP,
                    label="task_configuration",
                )
            )
        return task

    def assertConfigured(
        self,
        task: DBTask[Any, Any],
        task_configuration: Collection | None = None,
    ) -> None:
        """The task has been configured."""
        if task_configuration is not None:
            self.apply_configuration.assert_called_once_with(
                self.work_request.task_data, task_configuration, task
            )
        else:
            self.apply_configuration.assert_not_called()
        self.compute_scheduler_tags.assert_called_once()

        self.assertIsNotNone(self.work_request.configured_task_data)
        self.assertEqual(
            self.work_request.scheduler_tags_provided, sorted(self.provide_tags)
        )
        self.assertEqual(
            self.work_request.scheduler_tags_required, sorted(self.require_tags)
        )
        self.assertIsNone(self.work_request.dynamic_task_data)

    def test_errors_unset_configured_values(self) -> None:
        task_configuration = self.scenario.default_task_configuration_collection
        self.work_request.configured_task_data = {
            **self.sample_task_data,
            "build_architecture": "arm64",
        }
        self.work_request.dynamic_task_data = {
            **self.sample_dynamic_data,
            "input_source_artifact_id": 1001,
        }
        self.work_request.scheduler_tags_provided = ["provided"]
        self.work_request.scheduler_tags_required = ["required"]

        task = self.get_task(task_configuration=task_configuration)
        self.apply_configuration.side_effect = RuntimeError("expected error")
        with self.assertRaisesRegex(RuntimeError, "expected error"):
            task.apply_task_configuration()

        self.compute_scheduler_tags.assert_not_called()
        self.assertEqual(
            self.work_request.configured_task_data, self.work_request.task_data
        )
        self.assertEqual(self.work_request.scheduler_tags_provided, [])
        self.assertEqual(self.work_request.scheduler_tags_required, [])
        self.assertIsNone(self.work_request.dynamic_task_data)

    def test_config_collection_lookup_none(self) -> None:
        task = self.get_task(task_configuration=None)
        task.apply_task_configuration()
        self.assertConfigured(task)

    def test_config_collection_lookup_succeeds(self) -> None:
        task_configuration = self.scenario.default_task_configuration_collection
        task = self.get_task(task_configuration=task_configuration)
        task.apply_task_configuration()
        self.assertConfigured(task, task_configuration=task_configuration)

    def test_apply_configuration_fails(self) -> None:
        task = self.get_task()
        task.resolve_inputs(inputs.Stage.PENDING)
        self.apply_configuration.side_effect = TaskConfigError("expected error")
        with self.assertRaisesRegex(TaskConfigError, "expected error"):
            task.apply_task_configuration()

        self.assertEqual(
            self.work_request.configured_task_data,
            self.work_request.task_data,
        )
        self.assertIsNone(self.work_request.dynamic_task_data)
        self.assertEqual(self.work_request.scheduler_tags_provided, [])
        self.assertEqual(self.work_request.scheduler_tags_required, [])

    def test_apply_configuration_contributes_tags(self) -> None:
        self.work_request.save()
        task = self.work_request.get_task()
        self.apply_configuration.return_value = Changes(
            {}, {}, provide_tags={"p"}, require_tags={"r"}
        )
        tagset_add = self.enterContext(
            self.spy("debusine.server.tags.Tagset.add")
        )
        task.resolve_inputs(inputs.Stage.RUNNING)
        task.apply_task_configuration()

        self.assertEqual(
            self.work_request.configured_task_data,
            self.work_request.task_data,
        )
        self.assertIsNone(self.work_request.dynamic_task_data)

        # Tags have been contributed
        self.assertIn("p", self.work_request.scheduler_tags_provided)
        self.assertIn("r", self.work_request.scheduler_tags_required)

        # Tags have been contributed with the correct provenance
        p_found = False
        r_found = False
        for call in tagset_add.call_args_list:
            if call.args[2] == set("p"):
                p_found = True
                self.assertEqual(
                    call.args[1], ttags.ProvenanceProvided.WORKSPACE
                )
            elif call.args[2] == set("r"):
                r_found = True
                self.assertEqual(
                    call.args[1], wtags.ProvenanceRequired.WORKSPACE
                )
        self.assertTrue(p_found)
        self.assertTrue(r_found)

    def test_compute_scheduler_tags_fails(self) -> None:
        task = self.get_task()
        self.compute_scheduler_tags.side_effect = TaskConfigError(
            "expected error"
        )
        with self.assertRaisesRegex(TaskConfigError, "expected error"):
            task.apply_task_configuration()

        self.assertEqual(
            self.work_request.configured_task_data,
            self.work_request.task_data,
        )
        self.assertEqual(self.work_request.scheduler_tags_required, [])
        self.assertEqual(self.work_request.scheduler_tags_provided, [])
        self.assertIsNone(self.work_request.dynamic_task_data)

    def test_persist_input_fields(self) -> None:
        sid = self.playground.create_debian_environment(
            codename="sid", architecture="amd64"
        ).artifact
        assert sid is not None
        source = self.playground.create_minimal_source_package_artifact()
        work_request = self.playground.create_sbuild_work_request(
            source=source, environment=sid, architecture="amd64"
        )
        self.playground.advance_work_request(work_request, mark_running=True)
        assert work_request.dynamic_task_data is not None

        fields = work_request.dynamic_task_data["task_input_fields"]
        self.assertIn("input_source_artifact", fields)
        self.assertIn("input_extra_binary_artifacts", fields)
        self.assertIn("environment", fields)
        self.assertIn("extra_repositories", fields)
        self.assertIn("debusine_fqdn", fields)
        self.assertEqual(
            fields["input_source_artifact"]["artifact_id"], source.id
        )
        self.assertEqual(fields["environment"]["artifact_id"], sid.id)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to manage worker pools."""

import argparse
import sys
from collections.abc import Callable
from typing import Any, IO, NoReturn, cast, override

from django.core.exceptions import ValidationError
from django.core.management import CommandError, CommandParser
from django.db import transaction
from django.utils import timezone

from debusine.assets.models import AssetCategory
from debusine.db.models import Asset, Worker, WorkerPool
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.management.management_utils import WorkerPools
from debusine.server.worker_pools.models import WorkerPoolLimits
from debusine.utils.input import YamlEditor


class Command(DebusineBaseCommand):
    """Command to manage worker pools."""

    help = "Manage worker pools."

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        subparsers = parser.add_subparsers(dest="action", required=True)

        create = subparsers.add_parser(
            "create", help="Ensure a worker pool exists (create/update)"
        )
        create.add_argument("name", help="Name")
        create.add_argument(
            "--provider-account",
            type=self.get_provider_account,
            metavar="name",
            help="Name of cloud provider account",
        )
        create.add_argument(
            "--enabled",
            action=argparse.BooleanOptionalAction,
            default=True,
            help="Whether this worker pool is active (default: --enabled)",
        )
        create.add_argument(
            "--architectures",
            type=self.split_comma_separated_str,
            help=(
                "List of architectures supported by this pool (comma-separated)"
            ),
            required=True,
        )
        create.add_argument(
            "--tags",
            type=self.split_comma_separated_str,
            default=[],
            help="List of tags supported by this pool (comma-separated)",
        )
        create.add_argument(
            "--specifications",
            type=argparse.FileType("r"),
            help=(
                "File path (or - for stdin) to read the specification for the "
                "worker pool. YAML format. Defaults to stdin."
            ),
            default="-",
        )
        create.add_argument(
            "--instance-wide",
            action=argparse.BooleanOptionalAction,
            default=True,
            help=(
                "Whether this worker pool may be used by any scope "
                "(default: --instance-wide)"
            ),
        )
        create.add_argument(
            "--ephemeral",
            action=argparse.BooleanOptionalAction,
            default=False,
            help=(
                "Whether instances in this worker pool are terminated after "
                "every task (default: --no-ephemeral)"
            ),
        )
        create.add_argument(
            "--limits",
            type=argparse.FileType("r"),
            help=(
                "File path (or - for stdin) to read the limits for the "
                "worker pool. YAML format."
            ),
        )
        create.add_argument(
            "--apt-authentication",
            type=self.get_apt_authentication,
            metavar="name",
            help="Name of APT authentication asset.",
        )

        delete = subparsers.add_parser(
            "delete", help="Delete a worker pool and its workers"
        )
        delete.add_argument(
            "worker_pool", help="Name", type=self.get_worker_pool
        )
        delete.add_argument(
            "--force",
            action="store_true",
            help="Delete the worker pool even if it has active workers.",
        )

        edit = subparsers.add_parser(
            "edit", help="Edit a worker pool configuration"
        )
        edit.add_argument("worker_pool", help="Name", type=self.get_worker_pool)

        enable = subparsers.add_parser("enable", help="Enable a worker pool")
        enable.add_argument(
            "worker_pool", help="Name", type=self.get_worker_pool
        )

        disable = subparsers.add_parser("disable", help="Disable a worker pool")
        disable.add_argument(
            "worker_pool", help="Name", type=self.get_worker_pool
        )

        list_ = subparsers.add_parser("list", help="List worker pools")
        list_.add_argument(
            "--yaml", action="store_true", help="Machine readable YAML output"
        )

        launch = subparsers.add_parser("launch", help="Launch workers")
        launch.add_argument(
            "worker_pool", help="Name", type=self.get_worker_pool
        )
        launch.add_argument(
            "count", type=int, help="Number of workers to launch."
        )
        launch.add_argument(
            "--force",
            action="store_true",
            help="Launch workers, even if the pool is disabled",
        )

        terminate = subparsers.add_parser(
            "terminate", help="Terminate a pool worker"
        )
        terminate.add_argument(
            "worker", help="Worker Name", type=self.get_worker
        )

    @override
    def cleanup_arguments(self, *args: Any, **options: Any) -> None:
        for opt in ["specifications", "limits"]:
            if options.get(opt) and options[opt] != sys.stdin:
                options[opt].close()

    def get_provider_account(self, name: str) -> Asset:
        """Look up a cloud provider account asset by name."""
        try:
            return Asset.objects.get(
                category=AssetCategory.CLOUD_PROVIDER_ACCOUNT, data__name=name
            )
        except Asset.DoesNotExist:
            raise CommandError(
                f"Cloud provider account asset {name!r} not found", returncode=3
            )

    def get_apt_authentication(self, name: str) -> Asset:
        """Look up an APT authentication asset by name."""
        try:
            return Asset.objects.get_by_slug(
                category=AssetCategory.APT_AUTHENTICATION, slug=f":{name}"
            )
        except Asset.DoesNotExist:
            raise CommandError(
                f"APT authentication asset with slug ':{name}' not found",
                returncode=3,
            )

    def get_worker_pool(self, name: str) -> WorkerPool:
        """Look up a worker pool by name."""
        try:
            return WorkerPool.objects.get(name=name)
        except WorkerPool.DoesNotExist:
            raise CommandError(f"Worker Pool {name!r} not found", returncode=3)

    def get_worker(self, name: str) -> Worker:
        """Look up a worker by name."""
        try:
            return Worker.objects.get(name=name)
        except Worker.DoesNotExist:
            raise CommandError(f"Worker {name!r} not found", returncode=3)

    def split_comma_separated_str(self, combined: str) -> list[str]:
        """Split a comma-separated command line parameter."""
        return [item.strip() for item in combined.split(",")]

    def handle_create(
        self,
        *,
        name: str,
        provider_account: Asset,
        enabled: bool,
        architectures: list[str],
        tags: list[str] | None = None,
        specifications: IO[Any],
        instance_wide: bool,
        ephemeral: bool,
        limits: IO[Any] | None = None,
        apt_authentication: Asset | None = None,
        **options: Any,
    ) -> NoReturn:
        """
        Ensure a worker pool exists.

        This is idempotent, to make it easier to invoke from Ansible.
        """
        specifications_data = self.parse_yaml_data(specifications.read()) or {}
        limits_data: dict[str, Any] = WorkerPoolLimits().model_dump()
        if limits is not None:
            limits_data = self.parse_yaml_data(limits.read()) or {}
        with transaction.atomic():
            try:
                worker_pool = WorkerPool.objects.get(name=name)
            except WorkerPool.DoesNotExist:
                worker_pool = WorkerPool(
                    name=name, registered_at=timezone.now()
                )
            worker_pool.provider_account = provider_account
            worker_pool.enabled = enabled
            worker_pool.architectures = architectures
            worker_pool.tags = tags
            worker_pool.specifications = specifications_data
            worker_pool.instance_wide = instance_wide
            worker_pool.ephemeral = ephemeral
            worker_pool.limits = limits_data
            worker_pool.apt_authentication = apt_authentication
            try:
                worker_pool.full_clean()
                worker_pool.save()
            except ValidationError as exc:
                raise CommandError(
                    "Error creating worker_pool: " + "\n".join(exc.messages),
                    returncode=3,
                )

        raise SystemExit(0)

    def handle_delete(
        self, *, worker_pool: WorkerPool, force: bool, **options: Any
    ) -> NoReturn:
        """Delete the worker pool."""
        with transaction.atomic():
            if worker_pool.workers_running.exists() and not force:
                raise CommandError(
                    (
                        f"Worker pool {worker_pool.name!r} has "
                        f"{worker_pool.workers_running.count()} running "
                        f"workers."
                    ),
                    returncode=1,
                )

            for worker in worker_pool.workers_running:
                worker_pool.terminate_worker(worker)

            Worker.objects.filter(worker_pool=worker_pool).delete()
            worker_pool.delete()
        raise SystemExit(0)

    def handle_edit(
        self, *, worker_pool: WorkerPool, **options: Any
    ) -> NoReturn:
        """Edit the configuration of a worker pool."""
        data = {
            "provider_account": worker_pool.provider_account.data["name"],
            "apt_authentication": (
                worker_pool.apt_authentication.data["name"]
                if worker_pool.apt_authentication is not None
                else None
            ),
            "enabled": worker_pool.enabled,
            "architectures": worker_pool.architectures,
            "tags": worker_pool.tags,
            "specifications": worker_pool.specifications,
            "instance_wide": worker_pool.instance_wide,
            "ephemeral": worker_pool.ephemeral,
            "limits": worker_pool.limits,
        }
        editor: YamlEditor[dict[str, Any]] = YamlEditor(
            data,
            help_text=[
                "Here you can edit the configuration for the worker pool.",
                f"Worker pool: {worker_pool.name}",
            ],
        )
        if not editor.edit():
            self.stderr.write("Edit aborted.\n")
            raise SystemExit(1)
        if editor.value == data:
            self.stderr.write("No changes.\n")
            raise SystemExit(1)
        for k, old_v in data.items():
            if k not in editor.value:
                continue
            new_v = editor.value.get(k)
            if old_v != new_v:
                if k == "provider_account":
                    worker_pool.provider_account = Asset.objects.get(
                        category=AssetCategory.CLOUD_PROVIDER_ACCOUNT,
                        data__name=new_v,
                    )
                elif k == "apt_authentication":
                    apt_auth: Asset | None = None
                    if new_v is not None:
                        apt_auth = Asset.objects.get_by_slug(
                            category=AssetCategory.APT_AUTHENTICATION,
                            slug=f":{new_v}",
                        )
                    worker_pool.apt_authentication = apt_auth
                else:
                    setattr(worker_pool, k, new_v)
                self.stdout.write(f"Updated {k}.\n")
        worker_pool.save()
        raise SystemExit(0)

    def handle_enable(
        self, *, worker_pool: WorkerPool, **options: Any
    ) -> NoReturn:
        """Enable a worker pool."""
        worker_pool.enabled = True
        worker_pool.save()
        raise SystemExit(0)

    def handle_disable(
        self, *, worker_pool: WorkerPool, **options: Any
    ) -> NoReturn:
        """Disable a worker pool."""
        worker_pool.enabled = False
        worker_pool.save()
        raise SystemExit(0)

    def handle_list(
        self,
        yaml: bool,
        **options: Any,
    ) -> NoReturn:
        """List worker_pools."""
        worker_pools = WorkerPool.objects.all()
        WorkerPools(yaml).print(worker_pools, self.stdout)
        raise SystemExit(0)

    def handle_launch(
        self, *, worker_pool: WorkerPool, count: int, **options: Any
    ) -> NoReturn:
        """Launch workers in a pool."""
        worker_pool.launch_workers(count, override_disabled=options["force"])
        raise SystemExit(0)

    def handle_terminate(self, *, worker: Worker, **options: Any) -> NoReturn:
        """Terminate a worker in a pool."""
        if not worker.worker_pool:
            raise CommandError(
                f"Worker {worker.name!r} is not a member of a pool.",
                returncode=3,
            )
        worker.worker_pool.terminate_worker(worker)
        raise SystemExit(0)

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Dispatch the requested action."""
        func = cast(
            Callable[..., NoReturn],
            getattr(self, f"handle_{options['action']}", None),
        )
        if func is None:
            raise CommandError(
                f"Action {options['action']!r} not found", returncode=3
            )

        func(*args, **options)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: artifacts."""

import binascii
import logging
import tempfile
from collections.abc import Mapping
from contextlib import ExitStack
from pathlib import Path
from typing import Any, IO, Literal, override

from django.conf import settings
from django.core.exceptions import ValidationError as DjangoValidationError
from django.http import Http404
from django.utils import timezone
from rest_framework import serializers, status
from rest_framework.parsers import BaseParser, JSONParser
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import BaseSerializer

from debusine.artifacts.local_artifact import LocalArtifact
from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    File,
    FileInArtifact,
    FileUpload,
    User,
    WorkRequest,
)
from debusine.db.models.artifacts import (
    ArtifactQuerySet,
    ArtifactRelationQuerySet,
)
from debusine.db.models.permissions import anonymous_access
from debusine.server.exceptions import DebusineAPIException
from debusine.server.serializers import (
    ArtifactRelationResponseSerializer,
    ArtifactSerializer,
    ArtifactSerializerResponse,
)
from debusine.server.views.base import (
    BaseAPIView,
    CanDisplayFilterBackend,
    DestroyAPIViewBase,
    GenericAPIViewBase,
    GetOrCreateAPIView,
    ListAPIViewBase,
    RetrieveAPIViewBase,
)
from debusine.server.views.rest import (
    IsUserTokenAuthenticated,
    IsWorkRequestTokenAuthenticated,
)
from debusine.utils import parse_content_range_header

logger = logging.getLogger(__name__)


class ArtifactView(GenericAPIViewBase[Artifact], BaseAPIView):
    """View used to get or create artifacts."""

    filter_backends = [CanDisplayFilterBackend]
    lookup_url_kwarg = "artifact_id"

    @override
    def get_queryset(self) -> ArtifactQuerySet[Any]:
        """Get the query set for this view."""
        return Artifact.objects.in_current_scope()

    @override
    def get_object(self) -> Artifact:
        """Override to return more API-friendly errors."""
        try:
            return super().get_object()
        except Http404 as exc:
            raise DebusineAPIException(
                title=str(exc), status_code=status.HTTP_404_NOT_FOUND
            )

    def get(
        self,
        request: Request,
        artifact_id: int,  # noqa: ARG002, U100
    ) -> Response:
        """Return information of artifact_id."""
        artifact = self.get_object()
        self.set_current_workspace(artifact.workspace)
        self.enforce(artifact.can_display)

        artifact_response = ArtifactSerializerResponse.from_artifact(
            artifact, request
        )

        return Response(artifact_response.data, status=status.HTTP_200_OK)

    def _create_file_in_artifact(
        self, artifact: Artifact, path: str, file_deserialized: dict[str, Any]
    ) -> None:
        """Create a single file in an artifact."""
        file_size = file_deserialized["size"]
        hash_digest = binascii.unhexlify(
            file_deserialized["checksums"]["sha256"]
        )

        fileobj, created = File.objects.get_or_create(
            hash_digest=hash_digest,
            size=file_size,
        )

        file_in_artifact = FileInArtifact.objects.create(
            artifact=artifact,
            file=fileobj,
            path=path,
            complete=False,
            content_type=file_deserialized.get("content_type"),
        )

        if created or artifact.workspace.file_needs_upload(fileobj):
            if file_size == 0:
                # Create the file upload and add the file in the store
                with tempfile.NamedTemporaryFile(
                    prefix="file-uploading-empty-",
                    dir=settings.DEBUSINE_UPLOAD_DIRECTORY,
                    delete=False,
                ) as temp_empty_file:
                    pass
                file_upload = FileUpload.objects.create(
                    file_in_artifact=file_in_artifact,
                    path=Path(temp_empty_file.name).name,
                )
                UploadFileView.create_file_in_storage(file_upload)
                file_in_artifact.complete = True
                file_in_artifact.save()
        else:
            # The file already exists in the store and has a complete
            # FileInArtifact in the same workspace, indicating that it is
            # visible and we can safely copy a reference to it rather than
            # requiring it to be reuploaded.
            file_in_artifact.complete = True
            file_in_artifact.save()

    def post(self, request: Request) -> Response:
        """Create a new artifact."""
        data = request.data.copy()

        artifact_deserialized = ArtifactSerializer(data=data)
        artifact_deserialized.is_valid(raise_exception=True)

        workspace = artifact_deserialized.validated_data["workspace"]
        self.set_current_workspace(workspace)
        self.enforce(workspace.can_create_artifacts)

        work_request_id: int | None = artifact_deserialized.validated_data.get(
            "work_request"
        )
        work_request: WorkRequest | None = None
        created_by: User | None = None

        pc = context.pc
        if work_request_id is not None:
            if not pc.work_request_token:
                raise DebusineAPIException(
                    title=(
                        "work_request field can only be used by a request with "
                        "a token from a Worker"
                    ),
                    status_code=status.HTTP_403_FORBIDDEN,
                )

            try:
                work_request = WorkRequest.objects.get(id=work_request_id)
            except WorkRequest.DoesNotExist:
                raise DebusineAPIException(
                    title=f"WorkRequest {work_request_id} does not exist"
                )

            # TODO: do we allow users to create work requests with arbitrary
            # created_by_work_request values?
            if pc.work_request is not None and pc.work_request != work_request:
                raise DebusineAPIException(
                    title="Work request mismatch in POST data.",
                    status_code=status.HTTP_403_FORBIDDEN,
                )
        elif self.request.user.is_authenticated:
            created_by = self.request.user
        else:
            raise DebusineAPIException(
                title=(
                    "If work_request field is empty the WorkRequest must be "
                    "created using a Token associated to a user"
                ),
                status_code=status.HTTP_403_FORBIDDEN,
            )

        artifact = Artifact(
            category=artifact_deserialized["category"].value,
            workspace=workspace,
            data=artifact_deserialized["data"].value,
            created_by_work_request=work_request,
            expiration_delay=artifact_deserialized["expiration_delay"].value,
            created_by=created_by,
        )

        if artifact.category not in LocalArtifact.artifact_categories():
            raise DebusineAPIException(
                title="Error",
                detail=f"{artifact.category}: invalid artifact category",
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        try:
            artifact.full_clean()
        except DjangoValidationError as exc:
            raise serializers.ValidationError(exc.message_dict)

        artifact.save()

        for path, file_deserialized in artifact_deserialized[
            "files"
        ].value.items():
            self._create_file_in_artifact(artifact, path, file_deserialized)

        artifact_response = ArtifactSerializerResponse.from_artifact(
            artifact, request
        )
        return Response(artifact_response.data, status=status.HTTP_201_CREATED)


class WorkspaceArtifactView(RetrieveAPIViewBase[Artifact], BaseAPIView):
    """View used to get artifacts that may cross boundaries."""

    filter_backends = [CanDisplayFilterBackend]
    lookup_url_kwarg = "artifact_id"

    @override
    def get_queryset(self) -> ArtifactQuerySet[Any]:
        """Get the query set for this view."""
        return Artifact.objects.visible_to_current_workspace()

    @override
    def get_object(self) -> Artifact:
        """Override to return more API-friendly errors."""
        try:
            return super().get_object()
        except Http404 as exc:
            raise DebusineAPIException(
                title=str(exc), status_code=status.HTTP_404_NOT_FOUND
            )

    @override
    def get(
        self,
        request: Request,
        workspace: str,
        artifact_id: int,  # noqa: ARG002, U100
    ) -> Response:
        """Return information of artifact_id."""
        self.set_current_workspace(workspace)
        artifact = self.get_object()
        self.enforce(artifact.can_display)

        artifact_response = ArtifactSerializerResponse.from_artifact(
            artifact, request
        )

        return Response(artifact_response.data, status=status.HTTP_200_OK)


class FileUploadParser(BaseParser):
    """View used to upload files into an artifact."""

    media_type = "*/*"

    @staticmethod
    def _raise_error_if_gap_in_file(
        range_start: int, temporary_file_path: Path
    ) -> None:
        uploaded_size = Path(temporary_file_path).stat().st_size
        if uploaded_size < range_start:
            raise ValueError(
                f"Server had received {uploaded_size} bytes of the file. "
                f"New upload starts at {range_start}. "
                f"Please continue from {uploaded_size}"
            )

    # TODO: Can we avoid overriding the return type of BaseParser.parse?  It
    # doesn't seem to cause a practical problem, but is a bit odd.
    @override
    def parse(  # type: ignore[override]
        self,
        stream: IO[Any],
        media_type: str | None = None,
        parser_context: Mapping[str, Any] | None = None,
    ) -> FileUpload:
        """Upload a file to an artifact."""
        media_type  # fake usage for vulture
        assert parser_context is not None
        file_in_artifact = parser_context["request"].file_in_artifact

        range_start = 0
        range_end = None

        content_range = parse_content_range_header(
            parser_context["request"].headers
        )

        if content_range is not None:
            # Checked by UploadFileView.put.
            assert content_range["start"] != "*"

            range_start = content_range["start"]
            range_end = content_range["end"]

        if range_start > (file_size := file_in_artifact.file.size):
            raise ValueError(
                f"Range start ({range_start}) is greater than "
                f"file size ({file_size})"
            )

        with ExitStack() as stack:
            uploads_file: IO[bytes]
            try:
                file_upload = FileUpload.objects.get(
                    file_in_artifact=file_in_artifact
                )
                temporary_file_path = file_upload.absolute_file_path()
                self._raise_error_if_gap_in_file(
                    range_start, temporary_file_path
                )
                uploads_file = stack.enter_context(
                    open(temporary_file_path, "r+b")
                )
            except FileUpload.DoesNotExist:
                uploads_file = stack.enter_context(
                    tempfile.NamedTemporaryFile(
                        prefix='file-uploading-',
                        dir=settings.DEBUSINE_UPLOAD_DIRECTORY,
                        delete=False,
                        mode="a+b",
                    )
                )
                file_upload = FileUpload.objects.create(
                    file_in_artifact=file_in_artifact,
                    path=Path(uploads_file.name).name,
                )

            uploads_file.seek(range_start)

            bytes_received = 0
            # DATA_UPLOAD_MAX_MEMORY_SIZE: arbitrary size that should make
            # sense that can be dealt with
            while chunk := stream.read(settings.DATA_UPLOAD_MAX_MEMORY_SIZE):
                bytes_received += len(chunk)

                if range_start + bytes_received > file_size:
                    raise ValueError(
                        f"Cannot process range: attempted to write after end "
                        f"of the file. Range start: {range_start} "
                        f"Received: {bytes_received} File size: {file_size}"
                    )

                uploads_file.write(chunk)

            if (
                range_end is not None
                and (expected_bytes := range_end - range_start + 1)
                != bytes_received
            ):
                raise ValueError(
                    f"Expected {expected_bytes} bytes (based on range header: "
                    f"end-start+1). Received: {bytes_received} bytes"
                )

        file_upload.last_activity_at = timezone.now()
        file_upload.save()

        return file_upload


class UploadFileView(BaseAPIView):
    """View to upload files."""

    parser_classes = [FileUploadParser]
    # TODO: This should be replaced by appropriate debusine permissions.
    permission_classes = [
        IsUserTokenAuthenticated | IsWorkRequestTokenAuthenticated
    ]

    @staticmethod
    def _range_header(start: int, end: int) -> dict[str, str]:
        return {"Range": f"bytes={start}-{end}"}

    @classmethod
    def _file_status_response(
        cls,
        file_in_artifact: FileInArtifact,
        range_file_size: int | Literal["*"],
    ) -> Response:
        """
        Return response with the status of a file (completed or missing range).

        :param file_in_artifact: FileInArtifact that is being uploaded
        :param range_file_size: if it's "*" returns the total file size,
          if it contains a number and does not match the expected
          the server returns HTTP 400 Bad request
        """
        file_upload = getattr(file_in_artifact, "fileupload", None)

        if file_upload is None:
            return Response(
                status=status.HTTP_206_PARTIAL_CONTENT,
                headers=cls._range_header(0, 0),
            )

        file_in_db_size = file_upload.file_in_artifact.file.size
        if range_file_size != "*" and range_file_size != file_in_db_size:
            # Client reported a size in Content-Range that is not
            # what was expected
            raise DebusineAPIException(
                title=(
                    "Invalid file size in Content-Range header. "
                    f"Expected {file_in_db_size} "
                    f"received {range_file_size}"
                )
            )

        current_size = FileUpload.current_size(
            artifact=file_in_artifact.artifact,
            path_in_artifact=file_in_artifact.path,
        )

        if current_size == file_in_db_size:
            # No missing parts: file upload is already completed
            return Response(status=status.HTTP_200_OK)
        else:
            # Return HTTP 206 with the Range: 0-{uploaded_size}
            # so the client can continue the upload
            return Response(
                status=status.HTTP_206_PARTIAL_CONTENT,
                headers=cls._range_header(0, current_size),
            )

    def _get_file_in_artifact(
        self, artifact_id: int, file_path: str
    ) -> FileInArtifact:
        """Get file to be uploaded, or raise DebusineAPIException."""
        try:
            artifact = (
                Artifact.objects.in_current_scope()
                .can_display(context.pc)
                .get(id=artifact_id)
            )
        except Artifact.DoesNotExist:
            raise DebusineAPIException(
                title=f"Artifact {artifact_id} does not exist",
                status_code=status.HTTP_404_NOT_FOUND,
            )
        assert isinstance(artifact, Artifact)

        self.enforce(artifact.can_add_files)

        try:
            return FileInArtifact.objects.get(artifact=artifact, path=file_path)
        except FileInArtifact.DoesNotExist:
            raise DebusineAPIException(
                title=f'No file_path "{file_path}" for artifact {artifact.id}',
                status_code=status.HTTP_404_NOT_FOUND,
            )

    def put(
        self, request: Request, artifact_id: int, file_path: str
    ) -> Response:
        """Receive file to be uploaded or request uploaded content-range."""
        file_in_artifact = self._get_file_in_artifact(artifact_id, file_path)

        try:
            content_range = parse_content_range_header(request.headers)
        except ValueError as exc:
            raise DebusineAPIException(title=str(exc))

        artifact = file_in_artifact.artifact

        if content_range is not None and content_range["start"] == "*":
            # Client would like to know the status of a file
            # (which parts have been uploaded)
            assert content_range["size"] is not None
            return self._file_status_response(
                file_in_artifact, content_range["size"]
            )

        if file_in_artifact.complete:
            # File was already uploaded
            FileUpload.objects.filter(
                file_in_artifact=file_in_artifact
            ).delete()
            return Response(status=status.HTTP_201_CREATED)

        setattr(request, "file_in_artifact", file_in_artifact)
        # request.data use the FileUploadParser
        try:
            file_upload: FileUpload | dict[str, Any] = request.data
        except ValueError as value_error:
            raise DebusineAPIException(title=str(value_error))

        if file_upload == {}:
            # An empty POST will not have been parsed by DRF
            try:
                file_upload = FileUpload.objects.get(
                    file_in_artifact=file_in_artifact
                )
            except FileUpload.DoesNotExist:
                return Response(
                    status=status.HTTP_200_OK,
                    headers=self._range_header(0, 0),
                )
        assert isinstance(file_upload, FileUpload)

        current_size = FileUpload.current_size(
            artifact=artifact, path_in_artifact=file_path
        )

        if current_size != file_upload.file_in_artifact.file.size:
            return Response(
                status=status.HTTP_200_OK,
                headers=self._range_header(0, current_size),
            )
        else:
            # File is uploaded: create the file in the storage
            # Will return HTTP 201 Created (or HTTP 409 Conflict if the
            # expected hash does not match, the client will need to re-upload)
            try:
                return self.create_file_in_storage(file_upload)
            finally:
                file_upload.delete()

    @staticmethod
    def create_file_in_storage(file_upload: FileUpload) -> Response:
        """
        Add file_upload to the artifact's default store.

        Return Response: 201 Created if real hash and length with the
        expected hash and length match, 409 Conflict if no match
        (and deletes the FileUpload): the client need to upload the file again.
        """
        file_path = file_upload.absolute_file_path()

        hash_digest = File.calculate_hash(file_path)

        file_in_artifact = file_upload.file_in_artifact

        expected_file_hash_digest = file_in_artifact.file.hash_digest
        if file_in_artifact.file.hash_digest.hex() != hash_digest.hex():
            raise DebusineAPIException(
                title=(
                    "Invalid file hash. Expected: "
                    f"{expected_file_hash_digest.hex()} "
                    f"Received: {hash_digest.hex()}"
                ),
                status_code=status.HTTP_409_CONFLICT,
                # Don't roll back; the caller wants to delete the FileUpload
                # in this case, and we haven't made any other database
                # changes yet.
                rollback_transaction=False,
            )

        scope = file_in_artifact.artifact.workspace.scope
        file_backend = scope.upload_file_backend(file_in_artifact.file)

        file_backend.add_file(file_path, fileobj=file_in_artifact.file)
        file_in_artifact.complete = True
        file_in_artifact.save()

        return Response(status=status.HTTP_201_CREATED)


class ArtifactRelationsView(
    GetOrCreateAPIView[ArtifactRelation],
    DestroyAPIViewBase[ArtifactRelation],
    ListAPIViewBase[ArtifactRelation],
    BaseAPIView,
):
    """Return, create and delete relations between artifacts."""

    serializer_class = ArtifactRelationResponseSerializer
    filter_backends = [CanDisplayFilterBackend]
    parser_classes = [JSONParser]
    pagination_class = None

    @override
    def get_queryset(self) -> ArtifactRelationQuerySet[Any]:
        """Get the query set for this view."""
        # queryset is filtered to can_display
        self.enforce(anonymous_access)
        pc = context.pc
        queryset = ArtifactRelation.objects.in_current_scope().can_display(pc)
        filter_kwargs = {}

        if self.request.method == "GET":
            artifact_id = self.request.GET.get("artifact", None)
            target_artifact_id = self.request.GET.get("target_artifact", None)

            if artifact_id is not None:
                filter_kwargs = {"artifact_id": artifact_id}
                required_artifact_id = artifact_id
            elif target_artifact_id is not None:
                filter_kwargs = {"target_id": target_artifact_id}
                required_artifact_id = target_artifact_id
            else:
                raise DebusineAPIException(
                    title=(
                        '"artifact" or "target_artifact" parameter are '
                        'mandatory'
                    )
                )
            try:
                Artifact.objects.in_current_scope().can_display(pc).get(
                    id=required_artifact_id
                )
            except Artifact.DoesNotExist:
                raise DebusineAPIException(
                    title=f"Artifact {required_artifact_id} not found",
                    status_code=status.HTTP_404_NOT_FOUND,
                )
        return queryset.filter(**filter_kwargs)

    @override
    def perform_create(
        self, serializer: BaseSerializer[ArtifactRelation]
    ) -> None:
        """Create an artifact relation."""
        artifact = serializer.validated_data["artifact"]
        target = serializer.validated_data["target"]

        self.set_current_workspace(artifact.workspace)
        self.enforce(artifact.can_add_relation)
        self.enforce(target.can_display)
        # self.enforce(artifact.workspace.can_create_artifacts)

        if artifact.fileinartifact_set.filter(complete=False).exists():
            raise DebusineAPIException(
                f"Cannot create relation: source artifact {artifact.id} is "
                f"incomplete"
            )
        if target.fileinartifact_set.filter(complete=False).exists():
            raise DebusineAPIException(
                f"Cannot create relation: target artifact {target.id} is "
                f"incomplete"
            )
        super().perform_create(serializer)

    @override
    def perform_destroy(self, instance: ArtifactRelation) -> None:
        self.enforce(instance.can_delete)
        super().perform_destroy(instance)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine Cli Asset commands."""

import base64
import math
from typing import Any
from unittest import mock
from unittest.mock import MagicMock

import yaml

from debusine.assets import AssetCategory, KeyPurpose, SigningKeyData
from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.debusine import Debusine
from debusine.client.exceptions import DebusineError
from debusine.client.models import AssetResponse, AssetsResponse


class CliCreateAssetTests(BaseCliTests):
    """Tests for the Cli functionality related to creating assets."""

    def patch_asset_create(self) -> MagicMock:
        """Patch Debusine.asset_create."""
        patcher_asset_create = mock.patch.object(
            Debusine, "asset_create", autospec=True
        )
        mocked_asset_create = patcher_asset_create.start()

        self.addCleanup(patcher_asset_create.stop)
        return mocked_asset_create

    def assert_create_signing_asset(
        self,
        use_stdin: bool = True,
        simulate_failure: bool = False,
    ) -> None:
        """Create a debusine:signing-key asset using the CLI."""
        category = AssetCategory.SIGNING_KEY
        asset_id = 2
        workspace = "test"
        data = SigningKeyData(
            purpose=KeyPurpose.OPENPGP_UPLOAD,
            fingerprint="ABC123",
            public_key=base64.b64encode(
                b"-----BEGIN PGP PUBLIC KEY BLOCK-----\n\n"
                b"-----END PGP PUBLIC KEY BLOCK-----\n"
            ).decode(),
            description="A Description",
        )

        mocked_asset_create = self.patch_asset_create()

        assert_system_exit_code: int | None = None
        if simulate_failure:
            error_detail: dict[str, Any] = {
                "title": "Workspace not found",
                "detail": "Workspace test not found in scope debusine",
            }
            mocked_asset_create.side_effect = DebusineError(**error_detail)
        else:
            mocked_asset_create.return_value = AssetResponse(
                id=asset_id,
                data=data.model_dump(),
                category=category,
                workspace=workspace,
            )

        create_cli_options = [category, "--workspace", workspace]

        yaml_data = yaml.safe_dump(data.model_dump())
        if use_stdin:
            self.enterContext(self.patch_sys_stdin_read(yaml_data))
            create_cli_options.extend(["--data", "-"])
        else:
            data_file_temp = self.create_temporary_file(
                prefix="create-asset",
                contents=yaml_data.encode("UTF-8"),
            )
            create_cli_options.extend(["--data", str(data_file_temp)])

        cli = self.create_cli(["asset", "create", *create_cli_options])

        if simulate_failure:
            exception = self.assertShowsError(cli.execute)
        else:
            stderr, stdout = self.capture_output(
                cli.execute, assert_system_exit_code=assert_system_exit_code
            )

        mocked_asset_create.assert_called_once_with(
            mock.ANY, category=category, data=data, workspace=workspace
        )

        server_url = self.servers[self.default_server]["api-url"]

        if simulate_failure:
            self.assertDebusineError(exception, error_detail)
        else:
            expected_data = {
                "result": "success",
                "message": (
                    f"New asset created in {server_url} in workspace "
                    f"{workspace} with id {asset_id}."
                ),
            }
            expected = yaml.safe_dump(
                expected_data, sort_keys=False, width=math.inf
            )
            if use_stdin:
                expected = f"---\n{expected}"

            self.assertEqual(stdout, expected)
            self.assertEqual(stderr, "")

    def test_create_signing_asset_stdin(self) -> None:
        """Test create an asset with stdin YAML data."""
        self.assert_create_signing_asset(use_stdin=True)

    def test_create_signing_asset_file(self) -> None:
        """Test create an asset with YAML data in a file."""
        self.assert_create_signing_asset(use_stdin=False)

    def test_create_signing_asset_existing(self) -> None:
        """Test create an asset with YAML data in a file."""
        self.assert_create_signing_asset(simulate_failure=True)


class CliListAssetTests(BaseCliTests):
    """Tests for the cli functionality related to listing assets."""

    def patch_asset_list(self) -> MagicMock:
        """Patch Debusine.asset_list."""
        patcher_asset_list = mock.patch.object(
            Debusine, "asset_list", autospec=True
        )
        mocked_asset_list = patcher_asset_list.start()

        self.addCleanup(patcher_asset_list.stop)
        return mocked_asset_list

    def assert_list_signing_assets(
        self,
        args: list[str],
        expected_call: mock._Call,
        simulate_failure: bool = False,
    ) -> None:
        """List assets using the CLI."""
        mocked_asset_list = self.patch_asset_list()
        asset_response = AssetResponse(
            id=2,
            category=AssetCategory.SIGNING_KEY,
            workspace="test",
            data=SigningKeyData(
                purpose=KeyPurpose.OPENPGP_UPLOAD,
                fingerprint="ABC123",
                public_key=base64.b64encode(
                    b"-----BEGIN PGP PUBLIC KEY BLOCK-----\n\n"
                    b"-----END PGP PUBLIC KEY BLOCK-----\n"
                ).decode(),
                description="A Description",
            ).model_dump(),
            work_request=12,
        )
        cli = self.create_cli(["asset", "list", *args])

        if simulate_failure:
            error_details: dict[str, Any] = {
                "title": "Workspace not found",
                "detail": "Workspace test not found in scope debusine",
            }
            mocked_asset_list.side_effect = DebusineError(**error_details)
            exception = self.assertShowsError(cli.execute)
        else:
            mocked_asset_list.return_value = AssetsResponse([asset_response])
            stderr, stdout = self.capture_output(cli.execute)

        mocked_asset_list.assert_has_calls([expected_call])
        self.assertEqual(len(mocked_asset_list.mock_calls), 1)

        if simulate_failure:
            self.assertDebusineError(exception, error_details)
        else:
            expected_data: list[dict[str, Any]] | dict[str, Any] = [
                asset_response.model_dump()
            ]
            expected = yaml.safe_dump(
                expected_data, sort_keys=False, width=math.inf
            )

            self.assertEqual(stdout, expected)
            self.assertEqual(stderr, "")

    def test_list_assets_by_id(self) -> None:
        """Test list an asset by ID."""
        self.assert_list_signing_assets(
            args=["--id", "2"],
            expected_call=mock.call(
                mock.ANY, asset_id=2, work_request=None, workspace=None
            ),
        )

    def test_list_assets_by_workspace(self) -> None:
        """Test list an asset by workspace."""
        self.assert_list_signing_assets(
            args=["--workspace", "test"],
            expected_call=mock.call(
                mock.ANY, asset_id=None, work_request=None, workspace="test"
            ),
        )

    def test_list_assets_by_work_request(self) -> None:
        """Test list an asset by work_request."""
        self.assert_list_signing_assets(
            args=["--work-request", "12"],
            expected_call=mock.call(
                mock.ANY, asset_id=None, work_request=12, workspace=None
            ),
        )

    def test_list_assets_failure(self) -> None:
        """Test list an asset by work_request."""
        self.assert_list_signing_assets(
            args=["--work-request", "12"],
            expected_call=mock.call(
                mock.ANY, asset_id=None, work_request=12, workspace=None
            ),
            simulate_failure=True,
        )

    def test_list_assets_unfiltered(self) -> None:
        cli = self.create_cli(["asset", "list"])

        stderr, _ = self.capture_output(cli.execute, assert_system_exit_code=3)
        self.assertEqual(
            stderr,
            (
                "At least one of --id, --workspace, and --work-request must "
                "be specified.\n"
            ),
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the workflow_utils functions."""

import re

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebianSourcePackage,
    DebusinePromise,
)
from debusine.client.models import LookupChildType
from debusine.db.models import ArtifactRelation, CollectionItem, WorkRequest
from debusine.db.models.lookups import (
    LookupResultArtifact,
    LookupResultBare,
    LookupResultCollection,
)
from debusine.server.workflows import (
    LintianWorkflow,
    NoopWorkflow,
    SbuildWorkflow,
    workflow_utils,
)
from debusine.server.workflows.models import (
    LintianWorkflowData,
    SbuildWorkflowData,
)
from debusine.tasks import TaskConfigError
from debusine.tasks.models import (
    BackendType,
    ExtraExternalRepository,
    ExtraRepository,
    LookupMultiple,
    SbuildInput,
)
from debusine.test.django import TestCase


class WorkflowUtilsTests(TestCase):
    """Tests for functions in workflows.workflow_utils."""

    @staticmethod
    def get_noop_workflow(work_request: WorkRequest) -> NoopWorkflow:
        """Return NoopWorkflow from work_request."""
        workflow = work_request.get_task()
        assert isinstance(workflow, NoopWorkflow)
        return workflow

    def test_source_package_input_and_source_package_data(self) -> None:
        source_package = self.playground.create_source_artifact(
            name="hello", version="1.0"
        )
        self.playground.create_debian_environment(
            vendor="debian", codename="trixie"
        )

        work_request = self.playground.create_workflow(
            "sbuild",
            SbuildWorkflowData(
                input=SbuildInput(source_artifact=source_package.id),
                target_distribution="debian:trixie",
                architectures=["amd64"],
            ),
        )

        workflow = work_request.get_task()
        assert isinstance(workflow, SbuildWorkflow)

        artifact = workflow_utils.source_package(
            workflow, configuration_key="input.source_artifact"
        )
        self.assertEqual(artifact, source_package)

        source_package_data = workflow_utils.source_package_data(
            workflow, configuration_key="input.source_artifact"
        )
        self.assertEqual(source_package_data.name, "hello")
        self.assertEqual(source_package_data.version, "1.0")

    def test_source_package_source_artifact(self) -> None:
        source_package = self.playground.create_source_artifact(
            name="hello", version="1.0"
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact()
        )

        self.playground.create_debian_environment(
            vendor="debian", codename="trixie"
        )

        work_request = self.playground.create_workflow(
            "lintian",
            LintianWorkflowData(
                source_artifact=source_package.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="trixie",
            ),
        )

        workflow = work_request.get_task()
        assert isinstance(workflow, LintianWorkflow)

        source_package_data = workflow_utils.source_package_data(workflow)

        self.assertEqual(source_package_data.name, "hello")
        self.assertEqual(source_package_data.version, "1.0")

    def test_lookup_result_artifact_category_artifact(self) -> None:
        """lookup_result_artifact_category with an artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={
                "name": "hello",
                "version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {"Architecture": "any"},
            },
        )
        result = LookupResultArtifact(lookup=artifact.id, artifact=artifact)

        self.assertEqual(
            workflow_utils.lookup_result_artifact_category(result),
            ArtifactCategory.SOURCE_PACKAGE,
        )

    def test_lookup_result_artifact_category_promise(self) -> None:
        """lookup_result_artifact_category with a promise."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        promise = self.playground.create_bare_data_item(
            collection,
            "test",
            category=BareDataCategory.PROMISE,
            data={
                "promise_work_request_id": 2,
                "promise_workflow_id": 1,
                "promise_category": ArtifactCategory.TEST,
            },
        )
        result = LookupResultBare(
            lookup=f"test@{CollectionCategory.TEST}/test",
            collection_item=promise,
        )

        self.assertEqual(
            workflow_utils.lookup_result_artifact_category(result),
            ArtifactCategory.TEST,
        )

    def test_lookup_result_artifact_category_other_bare_item(self) -> None:
        """lookup_result_artifact_category with a non-promise bare item."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        promise = self.playground.create_bare_data_item(
            collection, "test", category=BareDataCategory.TEST
        )
        result = LookupResultBare(
            lookup=f"test@{CollectionCategory.TEST}/test",
            collection_item=promise,
        )

        with self.assertRaisesRegex(
            ValueError,
            re.escape(
                f"Cannot determine artifact category for lookup result: "
                f"{result}"
            ),
        ):
            workflow_utils.lookup_result_artifact_category(result)

    def test_lookup_result_artifact_category_collection(self) -> None:
        """lookup_result_artifact_category raises ValueError for collections."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        result = LookupResultCollection(
            lookup=f"test@{CollectionCategory.TEST}", collection=collection
        )

        with self.assertRaisesRegex(
            ValueError,
            re.escape(
                f"Cannot determine artifact category for lookup result: "
                f"{result}"
            ),
        ):
            workflow_utils.lookup_result_artifact_category(result)

    def test_lookup_result_architectures_promise_with_architecture(
        self,
    ) -> None:
        """lookup_result_architectures with a binary promise."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        promise = self.playground.create_bare_data_item(
            collection,
            "test",
            category=BareDataCategory.PROMISE,
            data={
                "promise_work_request_id": 2,
                "promise_workflow_id": 1,
                "promise_category": ArtifactCategory.TEST,
                "architecture": "amd64",
            },
        )
        result = LookupResultBare(
            lookup=f"test@{CollectionCategory.TEST}/test",
            collection_item=promise,
        )

        self.assertEqual(
            workflow_utils.lookup_result_architectures(result), {"amd64"}
        )

    def test_lookup_result_architectures_promise_no_architecture(self) -> None:
        """lookup_result_architectures rejects promise without architecture."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        promise = self.playground.create_bare_data_item(
            collection,
            "test",
            category=BareDataCategory.PROMISE,
            data=DebusinePromise(
                promise_work_request_id=2,
                promise_workflow_id=1,
                promise_category=ArtifactCategory.TEST,
            ),
        )
        result = LookupResultBare(
            lookup=f"test@{CollectionCategory.TEST}/test",
            collection_item=promise,
        )

        with self.assertRaisesRegex(
            ValueError,
            re.escape(
                f"Cannot determine architecture for lookup result: {result}"
            ),
        ):
            workflow_utils.lookup_result_architectures(result)

    def test_lookup_result_architectures_binary_packages_artifact(self) -> None:
        """lookup_result_architectures with a binary-packages artifact."""
        artifact = self.playground.create_minimal_binary_package_artifact(
            architecture="i386",
        )
        result = LookupResultArtifact(lookup=artifact.id, artifact=artifact)

        self.assertEqual(
            workflow_utils.lookup_result_architectures(result), {"i386"}
        )

    def test_lookup_result_architectures_binary_packages_artifact_item(
        self,
    ) -> None:
        """lookup_result_architectures with a binary-packages artifact item."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        artifact = self.playground.create_minimal_binary_package_artifact(
            architecture="i386",
        )
        item = CollectionItem.objects.create_from_artifact(
            artifact,
            parent_collection=collection,
            name="hello",
            data={},
            created_by_user=self.playground.get_default_user(),
        )
        result = LookupResultArtifact(
            lookup=f"test@{CollectionCategory.TEST}/hello",
            collection_item=item,
            artifact=artifact,
        )

        self.assertEqual(
            workflow_utils.lookup_result_architectures(result), {"i386"}
        )

    def test_lookup_result_architectures_debian_binary_package(self) -> None:
        """lookup_result_architectures return arch from a binary package."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data={
                "srcpkg_name": "hello",
                "srcpkg_version": "1.0.0",
                "deb_fields": {"Architecture": "amd64"},
                "deb_control_files": [],
            },
        )

        result = LookupResultArtifact(lookup=artifact.id, artifact=artifact)

        self.assertEqual(
            workflow_utils.lookup_result_architectures(result),
            {"amd64"},
        )

    def test_lookup_result_architectures_debian_binary_package_item(
        self,
    ) -> None:
        """lookup_result_architectures returns arch from a binpkg item."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data={
                "srcpkg_name": "hello",
                "srcpkg_version": "1.0.0",
                "deb_fields": {"Architecture": "amd64"},
                "deb_control_files": [],
            },
        )
        item = CollectionItem.objects.create_from_artifact(
            artifact,
            parent_collection=collection,
            name="hello",
            data={},
            created_by_user=self.playground.get_default_user(),
        )
        result = LookupResultArtifact(
            lookup=f"test@{CollectionCategory.TEST}/hello",
            collection_item=item,
            artifact=artifact,
        )

        self.assertEqual(
            workflow_utils.lookup_result_architectures(result),
            {"amd64"},
        )

    def test_lookup_result_architectures_debian_upload(self) -> None:
        """lookup_result_architectures return arch from an upload artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.UPLOAD,
            data={
                "type": "dpkg",
                "changes_fields": {
                    "Architecture": "amd64",
                    "Files": [{"name": "test.deb"}],
                },
            },
        )

        result = LookupResultArtifact(lookup=artifact.id, artifact=artifact)

        self.assertEqual(
            workflow_utils.lookup_result_architectures(result),
            {"amd64"},
        )

    def test_lookup_result_architectures_mixed_upload(self) -> None:
        """lookup_result_architectures return arch from a mixed upload."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.UPLOAD,
            data={
                "type": "dpkg",
                "changes_fields": {
                    "Architecture": "amd64 i386 source",
                    "Files": [
                        {"name": "test.dsc"},
                        {"name": "test_amd64.deb"},
                        {"name": "test_i386.deb"},
                    ],
                },
            },
        )

        result = LookupResultArtifact(lookup=artifact.id, artifact=artifact)

        self.assertEqual(
            workflow_utils.lookup_result_architectures(result),
            {"amd64", "i386"},
        )

    def test_lookup_result_architectures_debian_package_build_log(self) -> None:
        """lookup_result_architectures return arch from a build log."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            data={
                "source": "hello",
                "version": "1.0.0-1",
                "architecture": "amd64",
                "filename": "hello.build",
            },
        )

        result = LookupResultArtifact(lookup=artifact.id, artifact=artifact)

        self.assertEqual(
            workflow_utils.lookup_result_architectures(result),
            {"amd64"},
        )

    def test_lookup_result_architectures_debian_upload_item(self) -> None:
        """lookup_result_architectures return arch from an upload item."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.UPLOAD,
            data={
                "type": "dpkg",
                "changes_fields": {
                    "Architecture": "amd64",
                    "Files": [{"name": "test.deb"}],
                },
            },
        )
        item = CollectionItem.objects.create_from_artifact(
            artifact,
            parent_collection=collection,
            name="hello",
            data={},
            created_by_user=self.playground.get_default_user(),
        )
        result = LookupResultArtifact(
            lookup=f"test@{CollectionCategory.TEST}/hello",
            collection_item=item,
            artifact=artifact,
        )

        self.assertEqual(
            workflow_utils.lookup_result_architectures(result),
            {"amd64"},
        )

    def test_lookup_result_architectures_other_artifacts(self) -> None:
        """lookup_result_architectures with source package artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={
                "name": "hello",
                "version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {"Architecture": "any"},
            },
        )
        result = LookupResultArtifact(lookup=artifact.id, artifact=artifact)

        with self.assertRaisesRegex(
            ValueError, f"Unexpected type: {DebianSourcePackage.__name__}"
        ):
            workflow_utils.lookup_result_architectures(result)

    def test_lookup_result_architectures_collection(self) -> None:
        """lookup_result_architectures with a collection."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        result = LookupResultCollection(
            lookup=f"test@{CollectionCategory.TEST}", collection=collection
        )

        with self.assertRaisesRegex(
            ValueError,
            re.escape(
                "Unexpected result: must have collection_item or artifact"
            ),
        ):
            workflow_utils.lookup_result_architectures(result)

    def test_lookup_result_architectures_collection_item_invalid(self) -> None:
        """lookup_result_architectures raises invalid architecture type."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        promise = self.playground.create_bare_data_item(
            collection,
            "test",
            category=BareDataCategory.PROMISE,
            data={
                "promise_work_request_id": 2,
                "promise_workflow_id": 1,
                "promise_category": ArtifactCategory.TEST,
                "architecture": 000,
            },
        )
        result = LookupResultBare(
            lookup=f"test@{CollectionCategory.TEST}/test",
            collection_item=promise,
        )

        with self.assertRaisesRegex(
            ValueError,
            "Cannot determine architecture for lookup result:",
        ):
            workflow_utils.lookup_result_architectures(result)

    def test_lookup_result_architectures_unexpected_result(self) -> None:
        """lookup_result_architectures raise error: unexpected result."""
        result = LookupResultBare(lookup="")

        with self.assertRaisesRegex(ValueError, "^Unexpected result: .*"):
            workflow_utils.lookup_result_architectures(result)

    def test_lookup_result_binary_package_name_promise_with_bpn(self) -> None:
        """lookup_result_binary_package_name with a binary promise."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        promise = self.playground.create_bare_data_item(
            collection,
            "test",
            category=BareDataCategory.PROMISE,
            data={
                "promise_work_request_id": 2,
                "promise_workflow_id": 1,
                "promise_category": ArtifactCategory.TEST,
                "binary_package_name": "hello",
            },
        )
        result = LookupResultBare(
            lookup=f"test@{CollectionCategory.TEST}/test",
            collection_item=promise,
        )

        self.assertEqual(
            workflow_utils.lookup_result_binary_package_name(result), "hello"
        )

    def test_lookup_result_binary_package_name_promise_no_bpn(self) -> None:
        """lookup_result_binary_package_name rejects promise without BPN."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        promise = self.playground.create_bare_data_item(
            collection,
            "test",
            category=BareDataCategory.PROMISE,
            data=DebusinePromise(
                promise_work_request_id=2,
                promise_workflow_id=1,
                promise_category=ArtifactCategory.TEST,
            ),
        )
        result = LookupResultBare(
            lookup=f"test@{CollectionCategory.TEST}/test",
            collection_item=promise,
        )

        with self.assertRaisesRegex(
            ValueError,
            re.escape(
                f"Cannot determine binary package name for lookup result: "
                f"{result}"
            ),
        ):
            workflow_utils.lookup_result_binary_package_name(result)

    def test_lookup_result_binary_package_name_debian_binary_package(
        self,
    ) -> None:
        """lookup_result_binary_package_name return BPN from a binary."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data={
                "srcpkg_name": "hello",
                "srcpkg_version": "1.0.0",
                "deb_fields": {"Package": "hello-bin"},
                "deb_control_files": [],
            },
        )

        result = LookupResultArtifact(lookup=artifact.id, artifact=artifact)

        self.assertEqual(
            workflow_utils.lookup_result_binary_package_name(result),
            "hello-bin",
        )

    def test_lookup_result_binary_package_name_debian_binary_package_item(
        self,
    ) -> None:
        """lookup_result_binary_package_name return BPN from a binary item."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data={
                "srcpkg_name": "hello",
                "srcpkg_version": "1.0.0",
                "deb_fields": {"Package": "hello-bin"},
                "deb_control_files": [],
            },
        )
        item = CollectionItem.objects.create_from_artifact(
            artifact,
            parent_collection=collection,
            name="hello",
            data={},
            created_by_user=self.playground.get_default_user(),
        )
        result = LookupResultArtifact(
            lookup=f"test@{CollectionCategory.TEST}/hello",
            collection_item=item,
            artifact=artifact,
        )

        self.assertEqual(
            workflow_utils.lookup_result_binary_package_name(result),
            "hello-bin",
        )

    def test_lookup_result_binary_package_name_other_artifacts(self) -> None:
        """lookup_result_binary_package_name with source package artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={
                "name": "hello",
                "version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {"Package": "hello-bin"},
            },
        )
        result = LookupResultArtifact(lookup=artifact.id, artifact=artifact)

        with self.assertRaisesRegex(
            workflow_utils.ArtifactHasNoBinaryPackageName,
            re.escape(f"{DebianSourcePackage!r}"),
        ):
            workflow_utils.lookup_result_binary_package_name(result)

    def test_lookup_result_binary_package_name_collection(self) -> None:
        """lookup_result_binary_package_name with a collection."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        result = LookupResultCollection(
            lookup=f"test@{CollectionCategory.TEST}", collection=collection
        )

        with self.assertRaisesRegex(
            ValueError,
            re.escape(
                "Unexpected result: must have collection_item or artifact"
            ),
        ):
            workflow_utils.lookup_result_binary_package_name(result)

    def test_lookup_result_binary_package_name_collection_item_invalid(
        self,
    ) -> None:
        """lookup_result_binary_package_name raise error: invalid BPN type."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        promise = self.playground.create_bare_data_item(
            collection,
            "test",
            category=BareDataCategory.PROMISE,
            data={
                "promise_work_request_id": 2,
                "promise_workflow_id": 1,
                "promise_category": ArtifactCategory.TEST,
                "binary_package_name": 000,
            },
        )
        result = LookupResultBare(
            lookup=f"test@{CollectionCategory.TEST}/test",
            collection_item=promise,
        )

        with self.assertRaisesRegex(
            ValueError,
            "Cannot determine binary package name for lookup result:",
        ):
            workflow_utils.lookup_result_binary_package_name(result)

    def test_lookup_result_binary_package_name_unexpected_result(self) -> None:
        """lookup_result_binary_package_name raise error: unexpected result."""
        result = LookupResultBare(lookup="")

        with self.assertRaisesRegex(ValueError, "^Unexpected result: .*"):
            workflow_utils.lookup_result_binary_package_name(result)

    def test_filter_artifact_lookup_by_arch(self) -> None:
        wf = self.get_noop_workflow(
            self.playground.create_workflow(task_name="noop")
        )
        assert wf.work_request.internal_collection
        for architecture in ("amd64", "i386"):
            wf.work_request.internal_collection.manager.add_bare_data(
                name=f"hello-{architecture}",
                category=BareDataCategory.PROMISE,
                workflow=wf.work_request,
                data={
                    "promise_category": ArtifactCategory.BINARY_PACKAGE,
                    "promise_work_request_id": 42,
                    "promise_workflow_id": wf.work_request.id,
                    "architecture": architecture,
                    "binary_package_name": "hello",
                },
                user=self.playground.get_default_user(),
            )
        for architecture in ("armel", "armhf"):
            wf.work_request.internal_collection.manager.add_artifact(
                self.playground.create_minimal_binary_package_artifact(
                    architecture=architecture
                ),
                name=f"hello-{architecture}",
                variables={
                    "architecture": architecture,
                    "binary_package_name": "hello",
                },
                workflow=wf.work_request,
                user=self.playground.get_default_user(),
            )
        self.assertEqual(
            workflow_utils.filter_artifact_lookup_by_arch(
                wf,
                LookupMultiple.model_validate(
                    [
                        {
                            "collection": "internal@collections",
                            "data__binary_package_name": "hello",
                            "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        }
                    ]
                ),
                {"amd64", "armel", "ppc64el"},
            ),
            LookupMultiple(
                (
                    "internal@collections/name:hello-amd64",
                    "internal@collections/name:hello-armel",
                )
            ),
        )

    def test_filter_artifact_lookup_by_arch_with_upload(self) -> None:
        wf = self.get_noop_workflow(
            self.playground.create_workflow(task_name="noop")
        )
        artifacts = self.playground.create_upload_artifacts()
        architectures = {"amd64", "all"}
        self.assertEqual(
            workflow_utils.filter_artifact_lookup_by_arch(
                wf,
                LookupMultiple((f"{artifacts.upload.id}@artifacts",)),
                architectures,
            ),
            LookupMultiple((f"{artifacts.upload.id}@artifacts",)),
        )

    def test_filter_artifact_lookup_by_arch_with_mixed_upload(self) -> None:
        wf = self.get_noop_workflow(
            self.playground.create_workflow(task_name="noop")
        )
        artifacts = self.playground.create_upload_artifacts(
            binaries=[
                ("hello", "amd64"),
                ("hello", "i386"),
                ("hello-doc", "all"),
            ]
        )
        architectures = {"amd64", "all"}
        self.assertEqual(
            workflow_utils.filter_artifact_lookup_by_arch(
                wf,
                LookupMultiple((f"{artifacts.upload.id}@artifacts",)),
                architectures,
            ),
            LookupMultiple.model_validate(
                [
                    f"{artifact.id}@artifacts"
                    for artifact in artifacts.binaries
                    if artifact.data["deb_fields"]["Architecture"]
                    in architectures
                ]
            ),
        )

    def test_get_lookup_architectures(self) -> None:
        wf = self.get_noop_workflow(
            self.playground.create_workflow(task_name="noop")
        )
        assert wf.work_request.internal_collection
        wf.work_request.internal_collection.manager.add_bare_data(
            name="hello-amd64",
            category=BareDataCategory.PROMISE,
            workflow=wf.work_request,
            data={
                "promise_category": ArtifactCategory.BINARY_PACKAGE,
                "promise_work_request_id": 42,
                "promise_workflow_id": wf.work_request.id,
                "architecture": "amd64",
                "binary_package_name": "hello",
            },
            user=self.playground.get_default_user(),
        )
        wf.work_request.internal_collection.manager.add_artifact(
            self.playground.create_minimal_binary_package_artifact(
                architecture="i386"
            ),
            name="hello-i386",
            variables={
                "architecture": "i386",
                "binary_package_name": "hello",
            },
            workflow=wf.work_request,
            user=self.playground.get_default_user(),
        )
        self.assertEqual(
            workflow_utils.get_lookup_architectures(
                wf,
                LookupMultiple.model_validate(
                    [
                        {
                            "collection": "internal@collections",
                            "data__binary_package_name": "hello",
                            "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        }
                    ]
                ),
            ),
            {"amd64", "i386"},
        )

    def test_get_lookup_architectures_mixed_upload(self) -> None:
        wf = self.get_noop_workflow(
            self.playground.create_workflow(task_name="noop")
        )
        artifacts = self.playground.create_upload_artifacts(
            binaries=[
                ("hello", "amd64"),
                ("hello", "i386"),
                ("hello-doc", "all"),
            ],
        )
        self.assertEqual(
            workflow_utils.get_lookup_architectures(
                wf, LookupMultiple((f"{artifacts.upload.id}@artifacts",))
            ),
            {"amd64", "i386", "all"},
        )

    def test_locate_debian_source_package_with_srcpkg(self) -> None:
        """Test with a debian:source-package."""
        source_artifact = self.playground.create_source_artifact()
        result = workflow_utils.locate_debian_source_package(
            "source_artifact", source_artifact
        )
        self.assertEqual(result, source_artifact)

    def test_locate_debian_source_package_with_upload(self) -> None:
        """Test with a debian:source-package."""
        upload_artifacts = self.playground.create_upload_artifacts()
        result = workflow_utils.locate_debian_source_package(
            "upload_artifact", upload_artifacts.upload
        )
        self.assertEqual(result, upload_artifacts.source)

    def test_locate_debian_source_package_with_upload_without_relations(
        self,
    ) -> None:
        """Test with a lone debian:upload (no relations)."""
        upload_artifacts = self.playground.create_upload_artifacts(source=False)

        with self.assertRaisesRegex(
            TaskConfigError,
            r"Unable to find an artifact of category debian:source-package "
            r'with a relationship of type extends from ".* debian:upload .*"',
        ):
            workflow_utils.locate_debian_source_package(
                "upload_artifact",
                upload_artifacts.upload,
            )

    def test_locate_debian_source_package_with_upload_with_multiple_relations(
        self,
    ) -> None:
        """Test with a debian:upload extending multiple source packages."""
        upload_artifacts = self.playground.create_upload_artifacts(source=True)
        source2 = self.playground.create_source_artifact()
        self.playground.create_artifact_relation(
            artifact=upload_artifacts.upload,
            target=source2,
            relation_type=ArtifactRelation.Relations.EXTENDS,
        )

        with self.assertRaisesRegex(
            TaskConfigError,
            r"Multiple artifacts of category debian:source-package with "
            r'a relationship of type extends from ".* debian:upload .*" found',
        ):
            workflow_utils.locate_debian_source_package(
                "upload_artifact",
                upload_artifacts.upload,
            )

    def test_locate_debian_source_package_with_junk(self) -> None:
        """Test with a debian:source-package."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            r"^test_artifact: unexpected artifact category: "
            r"'debusine:test'. Valid categories: "
            r"\['debian:source-package', 'debian:upload'\]$",
        ):
            workflow_utils.locate_debian_source_package(
                "test_artifact", artifact
            )

    def test_locate_debian_binary_packages_with_binary_package(self) -> None:
        """Test with debian:binary-package artifacts."""
        binary_artifacts = [
            self.playground.create_minimal_binary_package_artifact()
            for _ in range(2)
        ]
        result = workflow_utils.locate_debian_binary_packages(
            "source_artifact", binary_artifacts
        )
        self.assertCountEqual(result, binary_artifacts)

    def test_locate_debian_binary_packages_with_upload(self) -> None:
        """Test with a debian:upload artifact."""
        upload_artifacts = self.playground.create_upload_artifacts(
            binaries=[("foo", "amd64"), ("foo-common", "all")]
        )
        result = workflow_utils.locate_debian_binary_packages(
            "source_artifact", [upload_artifacts.upload]
        )
        self.assertCountEqual(result, upload_artifacts.binaries)

    def test_locate_debian_binary_packages_with_upload_no_binaries(
        self,
    ) -> None:
        """Test with a debian:upload without binaries."""
        upload_artifacts = self.playground.create_upload_artifacts(binary=False)
        result = workflow_utils.locate_debian_binary_packages(
            "source_artifact", [upload_artifacts.upload]
        )
        self.assertEqual(result, [])

    def test_locate_debian_binary_packages_with_junk(self) -> None:
        """Test with a non-binary-package-related artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            r"^test_artifact\[0\]: unexpected artifact category: "
            r"'debusine:test'. Valid categories: "
            r"\['debian:binary-package', 'debian:upload'\]$",
        ):
            workflow_utils.locate_debian_binary_packages(
                "test_artifact", [artifact]
            )

    def test_follow_artifact_relations_multiple_found(self) -> None:
        """Test with a debian:upload extending multiple srcpkgs."""
        upload_artifacts = self.playground.create_upload_artifacts(source=True)
        source2 = self.playground.create_source_artifact()
        self.playground.create_artifact_relation(
            artifact=upload_artifacts.upload,
            target=source2,
            relation_type=ArtifactRelation.Relations.EXTENDS,
        )

        self.assertQuerySetEqual(
            workflow_utils.follow_artifact_relations(
                upload_artifacts.upload,
                ArtifactRelation.Relations.EXTENDS,
                ArtifactCategory.SOURCE_PACKAGE,
            ),
            [upload_artifacts.source, source2],
        )

    def test_follow_artifact_relations_not_found(self) -> None:
        """Test with a lone debian:upload (no relations)."""
        upload_artifacts = self.playground.create_upload_artifacts(source=False)

        self.assertQuerySetEqual(
            workflow_utils.follow_artifact_relations(
                upload_artifacts.upload,
                ArtifactRelation.Relations.EXTENDS,
                ArtifactCategory.SOURCE_PACKAGE,
            ),
            [],
        )

    def test_follow_artifact_relations_upload(self) -> None:
        """Test finding a source package from an upload."""
        upload_artifacts = self.playground.create_upload_artifacts()
        test_artifact, _ = self.playground.create_artifact()

        # Create some extra unrelated relations
        self.playground.create_artifact_relation(
            artifact=upload_artifacts.upload,
            target=upload_artifacts.binaries[0],
            relation_type=ArtifactRelation.Relations.BUILT_USING,
        )
        self.playground.create_artifact_relation(
            artifact=upload_artifacts.upload,
            target=test_artifact,
            relation_type=ArtifactRelation.Relations.EXTENDS,
        )

        self.assertQuerySetEqual(
            workflow_utils.follow_artifact_relations(
                upload_artifacts.upload,
                ArtifactRelation.Relations.EXTENDS,
                ArtifactCategory.SOURCE_PACKAGE,
            ),
            [upload_artifacts.source],
        )

    def test_locate_debian_source_package_lookup_srcpkg(self) -> None:
        """Test locate_debian_source_package_lookup() with a source package."""
        workflow = self.get_noop_workflow(
            self.playground.create_workflow(task_name="noop")
        )
        source_artifact = self.playground.create_source_artifact()
        self.assertEqual(
            workflow_utils.locate_debian_source_package_lookup(
                workflow, "source_artifact", source_artifact.id
            ),
            source_artifact.id,
        )

    def test_locate_debian_source_package_lookup_upload(self) -> None:
        """Test locate_debian_source_package_lookup() with a debian:upload."""
        workflow = self.get_noop_workflow(
            self.playground.create_workflow(task_name="noop")
        )
        artifacts = self.playground.create_upload_artifacts()
        self.assertEqual(
            workflow_utils.locate_debian_source_package_lookup(
                workflow, "source_artifact", artifacts.upload.id
            ),
            f"{artifacts.source.id}@artifacts",
        )

    def test_get_source_package_names_invalid_artifact_category(self) -> None:
        source_artifact = self.playground.create_source_artifact(name="hello")

        lookup_result = [
            LookupResultArtifact(
                artifact=source_artifact,
                lookup=source_artifact.id,
            )
        ]

        with self.assertRaisesRegex(
            TaskConfigError,
            r"^testing: unexpected artifact category: 'debian:source-package'. "
            r"Valid categories: \['debian:binary-package'\]$",
        ):
            workflow_utils.get_source_package_names(
                lookup_result,
                configuration_key="testing",
                artifact_expected_categories=(ArtifactCategory.BINARY_PACKAGE,),
            )

    def test_get_source_package_names_invalid_artifact_in_promise_category(
        self,
    ) -> None:
        lookup_result = [
            LookupResultBare(
                lookup=f"test@{CollectionCategory.TEST}/test",
                collection_item=CollectionItem(
                    data={
                        "promise_category": ArtifactCategory.TEST,
                        "source_package_name": "python3",
                    }
                ),
            ),
        ]

        with self.assertRaisesRegex(
            TaskConfigError,
            r"^testing: unexpected artifact category: 'debusine:test'. "
            r"Valid categories: \['debian:binary-package'\]$",
        ):
            workflow_utils.get_source_package_names(
                lookup_result,
                configuration_key="testing",
                artifact_expected_categories=(ArtifactCategory.BINARY_PACKAGE,),
            )

    def test_get_source_package_names(self) -> None:
        binary_artifact_1 = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello"
            )
        )
        binary_artifact_2 = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello"
            )
        )
        binaries_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="linux-base"
            )
        )
        upload_artifact = self.playground.create_upload_artifacts(
            src_name="firefox"
        ).upload

        lookup_result = [
            LookupResultArtifact(
                lookup=binary_artifact_1.id,
                artifact=binary_artifact_1,
            ),
            LookupResultArtifact(
                lookup=binary_artifact_2.id,
                artifact=binary_artifact_2,
            ),
            LookupResultArtifact(
                lookup=binaries_artifact.id,
                artifact=binaries_artifact,
            ),
            LookupResultArtifact(
                lookup=upload_artifact.id,
                artifact=upload_artifact,
            ),
            LookupResultBare(
                lookup=f"test@{CollectionCategory.TEST}/empty-binary",
                collection_item=CollectionItem(
                    # This isn't valid, without a source_package_name, but it
                    # gives coverage another path
                    data={
                        "promise_category": ArtifactCategory.BINARY_PACKAGE,
                    }
                ),
            ),
            LookupResultBare(
                lookup=f"test@{CollectionCategory.TEST}/python3",
                collection_item=CollectionItem(
                    data={
                        "promise_category": ArtifactCategory.BINARY_PACKAGE,
                        "source_package_name": "python3",
                    }
                ),
            ),
        ]

        self.assertEqual(
            workflow_utils.get_source_package_names(
                lookup_result,
                configuration_key="testing",
                artifact_expected_categories=(
                    ArtifactCategory.BINARY_PACKAGE,
                    ArtifactCategory.UPLOAD,
                ),
            ),
            ["firefox", "hello", "linux-base", "python3"],
        )

    def test_get_available_architectures(self) -> None:
        work_request = self.playground.create_workflow()
        workflow = self.get_noop_workflow(work_request)
        self.playground.create_debian_environment(
            codename="bookworm", architecture="amd64"
        )
        self.playground.create_debian_environment(
            codename="bookworm", architecture="i386"
        )
        self.assertEqual(
            workflow_utils.get_available_architectures(
                workflow, vendor="debian", codename="bookworm"
            ),
            {"amd64", "i386", "all"},
        )

    def test_get_available_architectures_no_environment(self) -> None:
        work_request = self.playground.create_workflow()
        workflow = self.get_noop_workflow(work_request)
        self.playground.create_debian_environments_collection()
        with self.assertRaisesRegex(
            TaskConfigError,
            r"Unable to find any environments for debian:bookworm",
        ):
            workflow_utils.get_available_architectures(
                workflow, vendor="debian", codename="bookworm"
            )

    def test_configure_for_overlay_suite_experimental(self) -> None:
        work_request = self.playground.create_workflow()
        workflow = self.get_noop_workflow(work_request)
        self.playground.create_debian_environment(
            codename="sid",
            variant="sbuild",
            mirror="http://ftp.uk.debian.org/debian",
            components=["main", "contrib"],
        )
        extra_repositories: list[ExtraRepository] | None
        for extra_repositories in (None, []):
            with self.subTest(extra_repositories=extra_repositories):
                result = workflow_utils.configure_for_overlay_suite(
                    workflow,
                    extra_repositories=extra_repositories,
                    vendor="debian",
                    codename="experimental",
                    environment="debian/match:codename=sid",
                    backend=BackendType.UNSHARE,
                    architecture="amd64",
                    try_variant="sbuild",
                )
                self.assertEqual(
                    result,
                    [
                        ExtraExternalRepository(
                            url="http://ftp.uk.debian.org/debian",
                            suite="experimental",
                            components=["main", "contrib"],
                        ),
                    ],
                )

    def test_configure_for_overlay_suite_no_overlay(self) -> None:
        work_request = self.playground.create_workflow()
        workflow = self.get_noop_workflow(work_request)
        extra_repositories = [
            ExtraExternalRepository(
                url="http://deb.debian.org/debian",
                suite="foo-backports",
                components=["main"],
            ),
        ]
        for codename in ("sid", "trixie", "bookworm-security"):
            with self.subTest(codename=codename):
                self.playground.create_debian_environment(codename=codename)
                self.assertEqual(
                    workflow_utils.configure_for_overlay_suite(
                        workflow,
                        extra_repositories=extra_repositories,
                        vendor="debian",
                        codename=codename,
                        environment=f"debian/match:codename={codename}",
                        backend=BackendType.UNSHARE,
                        architecture="amd64",
                        try_variant="autopkgtest",
                    ),
                    extra_repositories,
                )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the prometheus redis collector."""

from typing import override

from django.test import SimpleTestCase
from prometheus_client import values
from prometheus_client.core import CollectorRegistry, Sample
from prometheus_client.metrics import Counter, Gauge, Histogram

from debusine.server.prometheus_client.redis import (
    _daemon_threads,
    mark_process_dead,
    private_fake_redis,
    redis_client,
)
from debusine.server.prometheus_client.redis_collector import RedisCollector
from debusine.server.prometheus_client.values import RedisValue


@private_fake_redis()
class RedisCollectorTestCase(SimpleTestCase):
    """Test Redis collector."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.registry = CollectorRegistry()
        self.collector = RedisCollector(self.registry)
        redis_client().flushdb()

        old_values_class = values.ValueClass

        def restore_old_values_class() -> None:
            values.ValueClass = old_values_class

        values.ValueClass = RedisValue(lambda: 123)
        self.addCleanup(restore_old_values_class)

    @override
    def tearDown(self) -> None:
        super().tearDown()
        for identifier in list(_daemon_threads):
            mark_process_dead(identifier)

    def test_create_collector(self) -> None:
        collector = RedisCollector(None)
        self.assertEqual(list(collector.collect()), [])

    def test_counter(self) -> None:
        counter = Counter(
            "test",
            "Test Counter",
            labelnames=("label",),
            registry=self.registry,
        )
        counter.labels(label="a").inc()
        counter.labels(label="b").inc(7)

        metrics = {m.name: m for m in self.collector.collect()}
        self.assertEqual(
            metrics["test"].samples,
            [
                Sample("test_total", {"label": "a"}, 1.0),
                Sample("test_total", {"label": "b"}, 7.0),
            ],
        )

    def test_many_metrics(self) -> None:
        counter = Counter(
            "test",
            "Test Counter",
            labelnames=("label",),
            registry=self.registry,
        )
        for label in range(100):
            counter.labels(label=str(label)).inc()
        metrics = {m.name: m for m in self.collector.collect()}
        self.assertEqual(len(metrics["test"].samples), 100)

    def test_histogram(self) -> None:
        histogram = Histogram(
            "test",
            "Test Histogram",
            buckets=(0.01, 0.1, 1, 10, float("inf")),
            registry=self.registry,
        )
        histogram.observe(0.1)
        histogram.observe(0.5)
        histogram.observe(5)
        histogram.observe(50)
        metrics = {m.name: m for m in self.collector.collect()}
        self.assertEqual(
            metrics["test"].samples,
            [
                Sample("test_bucket", {"le": "0.01"}, 0.0),
                Sample("test_bucket", {"le": "0.1"}, 1.0),
                Sample("test_bucket", {"le": "1.0"}, 2.0),
                Sample("test_bucket", {"le": "10.0"}, 3.0),
                Sample("test_bucket", {"le": "+Inf"}, 4.0),
                Sample("test_count", {}, 4.0),
                Sample("test_sum", {}, 55.6),
            ],
        )

    def test_gauge_all(self) -> None:
        g1 = Gauge("g", "help", registry=None)
        values.ValueClass = RedisValue(lambda: 456)
        g2 = Gauge("g", "help", registry=None)
        self.assertEqual(0, self.registry.get_sample_value("g", {"pid": "123"}))
        self.assertEqual(0, self.registry.get_sample_value("g", {"pid": "456"}))
        g1.set(1)
        g2.set(2)
        mark_process_dead(123)
        self.assertEqual(1, self.registry.get_sample_value("g", {"pid": "123"}))
        self.assertEqual(2, self.registry.get_sample_value("g", {"pid": "456"}))

    def test_gauge_liveall(self) -> None:
        g1 = Gauge("g", "help", registry=None, multiprocess_mode="liveall")
        values.ValueClass = RedisValue(lambda: 456)
        g2 = Gauge("g", "help", registry=None, multiprocess_mode="liveall")
        self.assertEqual(0, self.registry.get_sample_value("g", {"pid": "123"}))
        self.assertEqual(0, self.registry.get_sample_value("g", {"pid": "456"}))
        g1.set(1)
        g2.set(2)
        self.assertEqual(1, self.registry.get_sample_value("g", {"pid": "123"}))
        self.assertEqual(2, self.registry.get_sample_value("g", {"pid": "456"}))
        mark_process_dead(123)
        self.assertEqual(
            None, self.registry.get_sample_value("g", {"pid": "123"})
        )
        self.assertEqual(2, self.registry.get_sample_value("g", {"pid": "456"}))

    def test_gauge_min(self) -> None:
        g1 = Gauge("g", "help", registry=None, multiprocess_mode="min")
        values.ValueClass = RedisValue(lambda: 456)
        g2 = Gauge("g", "help", registry=None, multiprocess_mode="min")
        self.assertEqual(0, self.registry.get_sample_value("g"))
        g1.set(1)
        g2.set(2)
        self.assertEqual(1, self.registry.get_sample_value("g"))

    def test_gauge_livemin(self) -> None:
        g1 = Gauge("g", "help", registry=None, multiprocess_mode="livemin")
        values.ValueClass = RedisValue(lambda: 456)
        g2 = Gauge("g", "help", registry=None, multiprocess_mode="livemin")
        self.assertEqual(0, self.registry.get_sample_value("g"))
        g1.set(1)
        g2.set(2)
        self.assertEqual(1, self.registry.get_sample_value("g"))
        mark_process_dead(123)
        self.assertEqual(2, self.registry.get_sample_value("g"))

    def test_gauge_max(self) -> None:
        g1 = Gauge("g", "help", registry=None, multiprocess_mode="max")
        values.ValueClass = RedisValue(lambda: 456)
        g2 = Gauge("g", "help", registry=None, multiprocess_mode="max")
        self.assertEqual(0, self.registry.get_sample_value("g"))
        g1.set(1)
        g2.set(2)
        self.assertEqual(2, self.registry.get_sample_value("g"))

    def test_gauge_livemax(self) -> None:
        g1 = Gauge("g", "help", registry=None, multiprocess_mode="livemax")
        values.ValueClass = RedisValue(lambda: 456)
        g2 = Gauge("g", "help", registry=None, multiprocess_mode="livemax")
        self.assertEqual(0, self.registry.get_sample_value("g"))
        g1.set(2)
        g2.set(1)
        self.assertEqual(2, self.registry.get_sample_value("g"))
        mark_process_dead(123)
        self.assertEqual(1, self.registry.get_sample_value("g"))

    def test_gauge_sum(self) -> None:
        g1 = Gauge("g", "help", registry=None, multiprocess_mode="sum")
        values.ValueClass = RedisValue(lambda: 456)
        g2 = Gauge("g", "help", registry=None, multiprocess_mode="sum")
        self.assertEqual(0, self.registry.get_sample_value("g"))
        g1.set(1)
        g2.set(2)
        self.assertEqual(3, self.registry.get_sample_value("g"))
        mark_process_dead(123)
        self.assertEqual(3, self.registry.get_sample_value("g"))

    def test_gauge_livesum(self) -> None:
        g1 = Gauge("g", "help", registry=None, multiprocess_mode="livesum")
        values.ValueClass = RedisValue(lambda: 456)
        g2 = Gauge("g", "help", registry=None, multiprocess_mode="livesum")
        self.assertEqual(0, self.registry.get_sample_value("g"))
        g1.set(1)
        g2.set(2)
        self.assertEqual(3, self.registry.get_sample_value("g"))
        mark_process_dead(123)
        self.assertEqual(2, self.registry.get_sample_value("g"))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Base support for tasks that run on server-side Celery workers."""

from abc import ABCMeta
from collections.abc import Callable
from typing import Any, override

from debusine.artifacts.models import TaskTypes
from debusine.db.context import context
from debusine.db.models import WorkRequest
from debusine.db.models.permissions import format_permission_check_error
from debusine.db.permissioncontext import PermissionContext
from debusine.db.tasks import DBTask
from debusine.tasks.models import BaseDynamicTaskData, BaseTaskData


class ServerTaskPermissionDenied(Exception):
    """Permission predicate checks failed on server tasks."""


def analyze_celery_worker_tasks() -> dict[str, Any]:
    """
    Return dictionary with metadata for each server task in DBTask._sub_tasks.

    Subclasses of DBTask get registered in DBTask._sub_tasks. Return
    a dictionary with the metadata of each of the subtasks.

    This method is executed in the worker when submitting the dynamic
    metadata.
    """
    metadata = {}
    registry = DBTask._sub_tasks.get(TaskTypes.SERVER, {})
    for task_class in registry.values():
        assert issubclass(task_class, BaseServerTask)
        metadata.update(task_class.analyze_worker())
    return metadata


def get_provided_celery_worker_tags() -> set[str]:
    """
    Return the set of worker tags provided by all available server tasks.

    This method is called on the Celery worker to collect information about
    what it can run.
    """
    tags: set[str] = set()
    registry = DBTask._sub_tasks.get(TaskTypes.SERVER, {})
    for task_class in registry.values():
        tags |= task_class.get_provided_worker_tags()
    return tags


class BaseServerTask[
    TD: BaseTaskData,
    DTD: BaseDynamicTaskData = BaseDynamicTaskData,
](DBTask[TD, DTD], metaclass=ABCMeta):
    """Base class for tasks that run on server-side Celery workers."""

    TASK_TYPE = TaskTypes.SERVER

    # If True, the task manages its own transactions.  If False, the task is
    # automatically run within a single transaction.
    TASK_MANAGES_TRANSACTIONS = False

    @override
    def get_task_configuration_subject_context(
        self,
    ) -> tuple[str | None, str | None]:
        return None, None

    @override
    def execute(self) -> WorkRequest.Results:
        """Execute task, setting up a suitable context."""
        context.reset()
        self.work_request.set_current()
        try:
            return super().execute()
        finally:
            context.reset()

    def enforce(self, predicate: Callable[[PermissionContext], bool]) -> None:
        """Enforce a permission predicate."""
        if predicate(
            PermissionContext.for_work_request(work_request=self.work_request)
        ):
            return

        raise ServerTaskPermissionDenied(
            format_permission_check_error(predicate, context.pc)
        )

    @classmethod
    def analyze_worker(cls) -> dict[str, Any]:
        """
        Return dynamic metadata about the current worker.

        This method is called on the worker to collect information about the
        worker. The information is stored as a set of key-value pairs in a
        dictionary.

        That information is then reused on the scheduler to be fed to
        :py:meth:`can_run_on` and determine if a task is suitable to be
        executed on the worker.

        Derived objects can extend the behaviour by overriding
        the method, calling ``metadata = super().analyze_worker()``,
        and then adding supplementary data in the dictionary.

        To avoid conflicts on the names of the keys used by different tasks
        you should use key names obtained with
        ``self.prefix_with_task_name(...)``.

        :return: a dictionary describing the worker.
        :rtype: dict.
        """
        version_key_name = cls.prefix_with_task_name("version")
        return {
            version_key_name: cls.TASK_VERSION,
        }



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Mixin to validate a sbuild build."""

from pathlib import Path, PurePath

from debian.deb822 import Deb822

from debusine.utils import (
    calculate_hash,
    find_file_suffixes,
    find_files_suffixes,
    read_changes,
    read_dsc,
)


class SbuildValidatorMixin:
    """
    Implement check_directory_for_consistency_errors for a Sbuild build.

    Encapsulate related methods.
    """

    def check_directory_for_consistency_errors(
        self, build_directory: Path
    ) -> list[str]:
        """
        Validate build_directory.

        Verify that no unexpected files in the directory, all the files
        in .changes exist in the directory and have the correct hash, etc.

        :return: list of errors (empty if no errors).
        """
        errors: list[str] = []

        files_in_directory: set[Path] = set(build_directory.glob("*"))

        # Files ending in .dsc, .changes or .build: exist in the directory
        # but are not taken in account for the checks: remove them from
        # files_in_directory
        remove_from_files_in_directory = find_files_suffixes(
            build_directory,
            [".dsc", ".changes", ".build"],
            include_symlinks=True,
        )

        for file in remove_from_files_in_directory:
            files_in_directory.discard(file)

        self._validate_changes_file_contents(
            build_directory, errors, files_in_directory
        )

        self._validate_dsc_file_contents(
            build_directory, errors, files_in_directory
        )

        # Check that no unexpected files exist in the directory
        # but not listed in the .changes or .dsc files
        if len(files_in_directory) > 0:
            files_in_directory_str = list(map(str, sorted(files_in_directory)))

            errors.append(
                f"Files in directory not referenced "
                f"in .changes or .dsc: {files_in_directory_str}"
            )

        return errors

    @classmethod
    def _validate_dsc_file_contents(
        cls,
        build_directory: Path,
        errors: list[str],
        files_in_directory: set[Path],
    ) -> None:
        dsc_path = find_file_suffixes(build_directory, [".dsc"])
        if dsc_path is None:
            return

        try:
            dsc = read_dsc(dsc_path)
        except ValueError as e:
            errors.append(str(e))
            return

        cls._validate_deb822_file(
            dsc, build_directory, errors, files_in_directory
        )

    @classmethod
    def _validate_changes_file_contents(
        cls,
        build_directory: Path,
        errors: list[str],
        files_in_directory: set[Path],
    ) -> None:
        changes_path = find_file_suffixes(build_directory, [".changes"])
        if changes_path is None:
            return

        try:
            changes = read_changes(changes_path)
        except ValueError as e:
            errors.append(str(e))
            return

        cls._validate_deb822_file(
            changes, build_directory, errors, files_in_directory
        )

    @staticmethod
    def _validate_deb822_file(
        contents: Deb822,
        build_directory: Path,
        errors: list[str],
        files_in_directory: set[Path],
    ) -> None:
        # hash algorithm to section
        hash_algorithm_to_section = {
            "sha1": "Checksums-Sha1",
            "sha256": "Checksums-Sha256",
            "md5": "Files",
        }
        # Each algorithm has a different field name when reading from
        # the deb822 file
        hash_algorithm_to_field = {
            "sha1": "sha1",
            "sha256": "sha256",
            "md5": "md5sum",
        }

        # It checks the file hashes for the given algorithms
        for hash_algorithm in ["sha1", "sha256", "md5"]:
            section_name = hash_algorithm_to_section[hash_algorithm]

            for file_in_changes in contents[section_name]:
                # Checked by read_changes.
                assert (
                    PurePath(file_in_changes["name"]).name
                    == file_in_changes["name"]
                )

                file = build_directory / file_in_changes["name"]

                if not file.is_file():
                    errors.append(
                        f"File in .changes section {section_name}"
                        f" does not exist: {file.name!r}"
                    )
                    continue

                field_name = hash_algorithm_to_field[hash_algorithm]
                expected_hash = file_in_changes[field_name]
                actual_hash = calculate_hash(file, hash_algorithm).hex()

                if expected_hash != actual_hash:
                    expected_size = file_in_changes["size"]
                    actual_size = file.stat().st_size
                    errors.append(
                        f"{hash_algorithm} "
                        f"for file {file.name} does not match: "
                        f"expected: {expected_hash} actual: {actual_hash}. "
                        f"Size expected: {expected_size} actual: {actual_size}"
                    )

                files_in_directory.discard(file)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for extracting signing input from other artifacts."""

import json
import logging
import shutil
import subprocess
from functools import partial
from pathlib import Path, PurePath
from typing import Any, AnyStr, override
from unittest import mock

import debusine.worker.tags as wtags
from debusine.artifacts import LocalArtifact, SigningInputArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    WorkRequestResults,
)
from debusine.artifacts.playground import ArtifactPlayground
from debusine.client.models import ArtifactResponse, RemoteArtifact
from debusine.tasks import ExtractForSigning, FALLBACK_ARCHITECTURE
from debusine.tasks.extract_for_signing import ExtractError
from debusine.tasks.models import (
    BaseTaskInputArtifact,
    ExtractForSigningDynamicData,
    WorkerType,
)
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_artifact_response,
    create_file_response,
    create_remote_artifact,
    create_system_tarball_data,
)


def _mock_run(
    args: list[str],
    run_as_root: bool = False,  # noqa: ARG001, U100
    **kwargs: Any,
) -> subprocess.CompletedProcess[AnyStr]:
    """Mock InstanceInterface.run using subprocess.run."""
    kwargs.setdefault("stdout", subprocess.PIPE)
    kwargs.setdefault("stderr", subprocess.PIPE)
    return subprocess.run(args, **kwargs)


class ExtractForSigningTests(
    ExternalTaskHelperMixin[ExtractForSigning], TestCase
):
    """Tests for ExtractForSigning."""

    SAMPLE_TASK_DATA = {
        "input": {"template_artifact": 2, "binary_artifacts": [3, 4]},
        "environment": "debian/match:codename=bookworm",
    }

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver.set_field_artifact(
            "environment",
            id=1,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )
        self.resolver.set_field_artifact(
            "template_artifact",
            id=2,
            category=ArtifactCategory.BINARY_PACKAGE,
            data=DebianBinaryPackage(
                srcpkg_name="template",
                srcpkg_version="1.0",
                deb_fields={"Package": "hello", "Version": "1.0-1"},
                deb_control_files=[],
            ),
        )
        self.resolver.set_field_artifacts(
            "binary_artifacts",
            BaseTaskInputArtifact(
                artifact_id=3,
                category=ArtifactCategory.BINARY_PACKAGE,
                data=DebianBinaryPackage(
                    srcpkg_name="unsigned",
                    srcpkg_version="1.0",
                    deb_fields={
                        "Package": "unsigned-1",
                        "Version": "2.0-1",
                        "Architecture": "amd64",
                    },
                    deb_control_files=[],
                ),
            ),
            BaseTaskInputArtifact(
                artifact_id=4,
                category=ArtifactCategory.BINARY_PACKAGE,
                data=DebianBinaryPackage(
                    srcpkg_name="unsigned",
                    srcpkg_version="1.0",
                    deb_fields={
                        "Package": "unsigned-2",
                        "Version": "3.0-1",
                        "Architecture": "arm64",
                    },
                    deb_control_files=[],
                ),
            ),
        )

    def test_can_run_on(self) -> None:
        """can_run_on returns True if unshare is available."""
        self.assertTrue(
            self.can_run_on(
                {
                    "system:worker_type": WorkerType.EXTERNAL,
                    "executor:unshare:available": True,
                    "extractforsigning:version": self.task.TASK_VERSION,
                    "system:architectures": [FALLBACK_ARCHITECTURE],
                }
            )
        )

    def test_can_run_on_mismatched_task_version(self) -> None:
        """can_run_on returns False for mismatched task versions."""
        self.assertFalse(
            self.can_run_on(
                {
                    "system:worker_type": WorkerType.EXTERNAL,
                    "executor:unshare:available": True,
                    "extractforsigning:version": self.task.TASK_VERSION + 1,
                }
            )
        )

    def test_can_run_on_unshare_not_available(self) -> None:
        """can_run_on returns False if unshare is not available."""
        self.assertFalse(
            self.can_run_on(
                {
                    "system:worker_type": WorkerType.EXTERNAL,
                    "executor:unshare:available": False,
                    "extractforsigning:version": self.task.TASK_VERSION,
                }
            )
        )

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:extractforsigning:version:1",
            ],
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            require_system={
                wtags.BUILD_ARCH_PREFIX + "amd64",
                wtags.EXECUTOR_PREFIX + "unshare",
            },
        )

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            ExtractForSigningDynamicData(
                environment_id=1,
                input_template_artifact_id=2,
                input_binary_artifacts_ids=[3, 4],
                subject="hello",
                parameter_summary="hello_1.0-1 (amd64, arm64)",
            ),
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "hello")

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_acceptable_input_categories(self) -> None:
        self.assertEqual(
            ExtractForSigning.inputs["template_artifact"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.BINARY_PACKAGE,),
        )
        self.assertEqual(
            ExtractForSigning.inputs["binary_artifacts"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.UPLOAD, ArtifactCategory.BINARY_PACKAGE),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = ExtractForSigningDynamicData(
            environment_id=1,
            input_template_artifact_id=2,
            input_binary_artifacts_ids=[3, 4],
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2, 3, 4])

    def test_fetch_input_template_wrong_category(self) -> None:
        """fetch_input checks the category of the template artifact."""
        self.task.dynamic_data = ExtractForSigningDynamicData(
            environment_id=1,
            input_template_artifact_id=2,
            input_binary_artifacts_ids=[3, 4],
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.download_artifact.return_value = create_artifact_response(
            id=2, category=ArtifactCategory.TEST
        )

        self.assertFalse(
            self.task.fetch_input(self.create_temporary_directory())
        )

        self.assertEqual(
            (self.task.log_directory / "fetch_input.log").read_text(),
            "Expected template_artifact to be of category "
            "debian:binary-package; got debusine:test\n",
        )

    def test_fetch_input_template_udeb(self) -> None:
        """fetch_input fails if the template does not have a .deb."""
        self.task.dynamic_data = ExtractForSigningDynamicData(
            environment_id=1,
            input_template_artifact_id=2,
            input_binary_artifacts_ids=[3, 4],
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.download_artifact.return_value = create_artifact_response(
            id=2,
            category=ArtifactCategory.BINARY_PACKAGE,
            data={"deb_fields": {"Package": "hello-signed-template"}},
            files={"hello-signed-template.udeb": create_file_response()},
        )

        self.assertFalse(
            self.task.fetch_input(self.create_temporary_directory())
        )

        self.assertEqual(
            (self.task.log_directory / "fetch_input.log").read_text(),
            "Expected template_artifact file name to match *.deb; "
            "got hello-signed-template.udeb\n",
        )

    def _create_binary_package_response(
        self, artifact_id: int, name: str, version: str
    ) -> ArtifactResponse:
        return create_artifact_response(
            id=artifact_id,
            category=ArtifactCategory.BINARY_PACKAGE,
            data={
                "srcpkg_name": name,
                "srcpkg_version": version,
                "deb_fields": {"Package": name},
                "deb_control_files": [],
            },
            files={f"{name}_{version}_amd64.deb": create_file_response()},
        )

    def test_fetch_input_gathers_information(self) -> None:
        """fetch_input gathers information from artifacts."""
        self.task.dynamic_data = ExtractForSigningDynamicData(
            environment_id=1,
            input_template_artifact_id=2,
            input_binary_artifacts_ids=[3, 4],
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.download_artifact.side_effect = [
            self._create_binary_package_response(
                2, "hello-signed-template", "1.0+1"
            ),
            self._create_binary_package_response(3, "hello", "1.0-1"),
            self._create_binary_package_response(4, "libhello1", "1.0-1"),
        ]
        destination = self.create_temporary_directory()

        self.assertTrue(self.task.fetch_input(destination))

        self.assertEqual(self.task._template_deb_name, "hello-signed-template")
        self.assertEqual(
            self.task._template_path,
            destination / "template" / "hello-signed-template_1.0+1_amd64.deb",
        )
        self.assertEqual(
            self.task._binary_artifacts, {"hello": 3, "libhello1": 4}
        )
        self.assertEqual(
            self.task._binary_paths,
            {
                "hello": destination / "binary" / "hello_1.0-1_amd64.deb",
                "libhello1": (
                    destination / "binary" / "libhello1_1.0-1_amd64.deb"
                ),
            },
        )

    def test_configure_for_execution(self) -> None:
        """configure_for_execution starts an executor instance."""
        download_directory = self.create_temporary_directory()
        mocked_prepare_executor_instance = (
            self.patch_prepare_executor_instance()
        )

        self.assertTrue(self.task.configure_for_execution(download_directory))

        mocked_prepare_executor_instance.assert_called_once_with()

    def test_pull_from_executor(self) -> None:
        """_pull_from_executor pulls a directory from the executor using tar."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        self.assertTrue(self.task.configure_for_execution(download_directory))
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.run.reset_mock()
        self.task.executor_instance.run.return_value = (
            subprocess.CompletedProcess([], 0)
        )

        self.task._remote_execute_directory = self.create_temporary_directory()
        self.task._local_execute_directory = self.create_temporary_directory()
        source = self.task._remote_execute_directory / "directory"
        target = self.task._local_execute_directory / "directory"
        source_tar = self.task._remote_execute_directory / "directory.tar"
        target_tar = self.task._local_execute_directory / "directory.tar"

        with mock.patch("subprocess.run") as mock_run:
            self.task._pull_from_executor(source, target)

        self.task.executor_instance.run.assert_called_once_with(
            ["tar", "-cf", f"{source}.tar", "directory"],
            run_as_root=False,
            cwd=self.task._remote_execute_directory,
            env=self.task._cmd_env(),
            stdout=mock.ANY,
            stderr=mock.ANY,
        )
        self.task.executor_instance.file_pull.assert_called_once_with(
            source_tar, target_tar
        )
        mock_run.assert_called_once_with(
            ["tar", "--one-top-level", "-xf", target_tar],
            cwd=self.task._local_execute_directory,
        )

    def test_extract_binary_error(self) -> None:
        """_extract_binary raises ExtractError if dpkg-deb exits non-zero."""
        execute_directory = self.create_temporary_directory()

        with (
            self.assertRaisesRegex(ExtractError, "dpkg-deb exited with code 1"),
            mock.patch.object(self.task, "run_cmd", return_value=1),
        ):
            self.task._extract_binary(
                Path("foo.deb"), execute_directory / "hello-signed-template"
            )

    def test_extract_binary_success(self) -> None:
        """_extract_binary calls dpkg-deb properly."""
        execute_directory = self.create_temporary_directory()

        with (
            self.assertLogs("debusine.tasks", level=logging.INFO) as log,
            mock.patch.object(
                self.task, "run_cmd", return_value=0
            ) as mock_run_cmd,
        ):
            self.task._extract_binary(
                Path("foo.deb"), execute_directory / "hello-signed-template"
            )

        mock_run_cmd.assert_called_once_with(
            [
                "dpkg-deb",
                "-x",
                "foo.deb",
                str(execute_directory / "hello-signed-template"),
            ],
            execute_directory,
        )
        self.assertEqual(
            log.output,
            [
                f"INFO:debusine.tasks:Executing: dpkg-deb -x foo.deb "
                f"{execute_directory}/hello-signed-template",
                "INFO:debusine.tasks:dpkg-deb exited with code 0",
            ],
        )

    def test_read_manifest(self) -> None:
        """_read_manifest extracts and parses the manifest."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        self.assertTrue(self.task.configure_for_execution(download_directory))
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.file_pull.side_effect = shutil.copy2
        self.task.executor_instance.run.side_effect = _mock_run
        self.task._remote_execute_directory = self.create_temporary_directory()
        self.task._local_execute_directory = self.create_temporary_directory()
        self.task._template_deb_name = "hello-signed-template"
        self.task._template_path = (
            download_directory
            / "template"
            / "hello-signed-template_1.0+1_amd64.deb"
        )
        manifest_path = PurePath(
            "usr/share/code-signing", self.task._template_deb_name, "files.json"
        )
        self.task._template_path.parent.mkdir()
        ArtifactPlayground.write_deb_file(
            self.task._template_path,
            data_files={manifest_path: json.dumps({"packages": {}}).encode()},
        )

        self.assertEqual(self.task._read_manifest(), {"packages": {}})

    def test_make_signing_input_artifact_invalid_package_name(self) -> None:
        """_make_signing_input_artifact fails if a package name is invalid."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        self.assertTrue(self.task.configure_for_execution(download_directory))
        self.task._binary_paths = {}
        self.task._remote_execute_directory = self.create_temporary_directory()
        self.task._local_execute_directory = self.create_temporary_directory()

        with self.assertRaisesRegex(
            ExtractError, "'_invalid' is not a valid package name"
        ):
            self.task._make_signing_input_artifact("_invalid", {})

    def test_make_signing_input_artifact_file_with_parent_segment(self) -> None:
        """_make_signing_input_artifact fails if a file name contains ".."."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        self.assertTrue(self.task.configure_for_execution(download_directory))
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.run.side_effect = _mock_run
        hello_path = download_directory / "binary" / "hello_1.0-1_amd64.deb"
        hello_path.parent.mkdir()
        ArtifactPlayground.write_deb_file(hello_path)
        self.task._binary_paths = {"hello": hello_path}
        self.task._remote_execute_directory = self.create_temporary_directory()
        self.task._local_execute_directory = self.create_temporary_directory()

        with (
            self.assertRaisesRegex(
                ExtractError,
                "File name 'usr/../../etc/passwd' may not contain '..' "
                "segments",
            ),
            mock.patch.object(
                self.task, "_pull_from_executor", side_effect=shutil.copytree
            ),
        ):
            self.task._make_signing_input_artifact(
                "hello",
                {
                    "files": [
                        {"sig_type": "efi", "file": "usr/../../etc/passwd"}
                    ]
                },
            )

    def test_make_signing_input_artifact_absolute_file(self) -> None:
        """_make_signing_input_artifact fails if a file name is absolute."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        self.assertTrue(self.task.configure_for_execution(download_directory))
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.run.side_effect = _mock_run
        hello_path = download_directory / "binary" / "hello_1.0-1_amd64.deb"
        hello_path.parent.mkdir()
        ArtifactPlayground.write_deb_file(hello_path)
        self.task._binary_paths = {"hello": hello_path}
        self.task._remote_execute_directory = self.create_temporary_directory()
        self.task._local_execute_directory = self.create_temporary_directory()

        with (
            self.assertRaisesRegex(
                ExtractError, "File name '/boot/vmlinuz' may not be absolute"
            ),
            mock.patch.object(
                self.task, "_pull_from_executor", side_effect=shutil.copytree
            ),
        ):
            self.task._make_signing_input_artifact(
                "hello",
                {"files": [{"sig_type": "efi", "file": "/boot/vmlinuz"}]},
            )

    def test_make_signing_input_artifact_symlink_escape(self) -> None:
        """_make_signing_input_artifact fails if a file escapes the tree."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        self.assertTrue(self.task.configure_for_execution(download_directory))
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.run.side_effect = _mock_run
        hello_path = download_directory / "binary" / "hello_1.0-1_amd64.deb"
        hello_path.parent.mkdir()
        ArtifactPlayground.write_deb_file(
            hello_path, data_symlinks={PurePath("evil-symlink"): PurePath("..")}
        )
        self.task._binary_paths = {"hello": hello_path}
        self.task._remote_execute_directory = self.create_temporary_directory()
        self.task._local_execute_directory = self.create_temporary_directory()

        with (
            self.assertRaisesRegex(
                ExtractError,
                "File name 'evil-symlink/etc/passwd' may not traverse symlinks "
                "to outside the package",
            ),
            mock.patch.object(
                self.task,
                "_pull_from_executor",
                side_effect=partial(shutil.copytree, symlinks=True),
            ),
        ):
            self.task._make_signing_input_artifact(
                "hello",
                {
                    "files": [
                        {"sig_type": "efi", "file": "evil-symlink/etc/passwd"}
                    ]
                },
            )

    def test_make_signing_input_artifact_valid(self) -> None:
        """_make_signing_input_artifact handles valid input."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        self.assertTrue(self.task.configure_for_execution(download_directory))
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.run.side_effect = _mock_run
        hello_path = download_directory / "binary" / "hello_1.0-1_amd64.deb"
        hello_path.parent.mkdir()
        ArtifactPlayground.write_deb_file(
            hello_path, data_files={PurePath("boot/vmlinuz"): b"vmlinuz"}
        )
        self.task._binary_paths = {"hello": hello_path}
        self.task._remote_execute_directory = self.create_temporary_directory()
        self.task._local_execute_directory = self.create_temporary_directory()

        with mock.patch.object(
            self.task, "_pull_from_executor", side_effect=shutil.copytree
        ):
            artifact = self.task._make_signing_input_artifact(
                "hello",
                {"files": [{"sig_type": "efi", "file": "boot/vmlinuz"}]},
            )

        self.assertEqual(
            artifact,
            SigningInputArtifact.create(
                [self.task._local_execute_directory / "hello/boot/vmlinuz"],
                self.task._local_execute_directory,
                binary_package_name="hello",
            ),
        )

    def test_make_signing_input_artifact_trusted_certs(self) -> None:
        """_make_signing_input_artifact passes through trusted_certs."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        self.assertTrue(self.task.configure_for_execution(download_directory))
        assert isinstance(self.task.executor_instance, mock.MagicMock)
        self.task.executor_instance.run.side_effect = _mock_run
        hello_path = download_directory / "binary" / "hello_1.0-1_amd64.deb"
        hello_path.parent.mkdir()
        ArtifactPlayground.write_deb_file(
            hello_path, data_files={PurePath("boot/vmlinuz"): b"vmlinuz"}
        )
        self.task._binary_paths = {"hello": hello_path}
        self.task._remote_execute_directory = self.create_temporary_directory()
        self.task._local_execute_directory = self.create_temporary_directory()

        with mock.patch.object(
            self.task, "_pull_from_executor", side_effect=shutil.copytree
        ):
            artifact = self.task._make_signing_input_artifact(
                "hello",
                {
                    "trusted_certs": ["0" * 64],
                    "files": [{"sig_type": "efi", "file": "boot/vmlinuz"}],
                },
            )

        self.assertEqual(
            artifact,
            SigningInputArtifact.create(
                [self.task._local_execute_directory / "hello/boot/vmlinuz"],
                self.task._local_execute_directory,
                trusted_certs=["0" * 64],
                binary_package_name="hello",
            ),
        )

    def test_upload_artifacts_execution_failed(self) -> None:
        """upload_artifacts does not upload anything if execution failed."""
        debusine_mock = self.mock_debusine()
        execute_directory = self.create_temporary_directory()

        self.task.upload_artifacts(
            execute_directory, execution_result=WorkRequestResults.FAILURE
        )

        debusine_mock.upload_artifact.assert_not_called()
        debusine_mock.relation_create.assert_not_called()

    def test_upload_artifacts_execution_succeeded(self) -> None:
        """upload_artifacts uploads if execution succeeded."""
        directory = self.create_temporary_directory()
        (directory / "hello/boot").mkdir(parents=True)
        (directory / "hello/boot/vmlinuz").write_bytes(b"vmlinuz")
        self.task.dynamic_data = ExtractForSigningDynamicData(
            environment_id=1,
            input_template_artifact_id=2,
            input_binary_artifacts_ids=[3, 4],
        )
        self.task._binary_artifacts = {"hello": 3}
        self.task._signing_input_artifacts = {
            "hello": SigningInputArtifact.create(
                [directory / "hello/boot/vmlinuz"], directory
            )
        }
        debusine_mock = self.mock_debusine()
        debusine_mock.upload_artifact.return_value = create_remote_artifact(
            id=5, workspace=self.task.workspace_name
        )
        execute_directory = self.create_temporary_directory()

        self.task.upload_artifacts(
            execute_directory, execution_result=WorkRequestResults.SUCCESS
        )

        debusine_mock.upload_artifact.assert_called_once_with(
            self.task._signing_input_artifacts["hello"],
            workspace=self.task.workspace_name,
            work_request=self.task.work_request_id,
        )
        debusine_mock.relation_create.assert_has_calls(
            [
                mock.call(
                    5,
                    self.task.dynamic_data.input_template_artifact_id,
                    "relates-to",
                ),
                mock.call(
                    5, self.task._binary_artifacts["hello"], "relates-to"
                ),
            ]
        )

    def test_cleanup_without_local_execute_directory(self) -> None:
        """cleanup() works even if _local_execute_directory hasn't been set."""
        # Just check that it can be called without raising an exception.
        self.task.cleanup()

    def test_execute_end_to_end(self) -> None:
        """End-to-end execution works."""
        self.task.dynamic_data = ExtractForSigningDynamicData(
            environment_id=1,
            input_template_artifact_id=2,
            input_binary_artifacts_ids=[3, 4],
        )

        # Create some test files.
        directory = self.create_temporary_directory()
        template_deb_path = directory / "hello-signed-template_1.0+1_amd64.deb"
        code_signing_path = PurePath(
            "usr/share/code-signing/hello-signed-template"
        )
        ArtifactPlayground.write_deb_file(
            template_deb_path,
            data_files={
                (code_signing_path / "files.json"): json.dumps(
                    {
                        "packages": {
                            "hello": {
                                "files": [
                                    {"sig_type": "efi", "file": "boot/vmlinuz"}
                                ]
                            },
                            "grub-efi-amd64-bin": {
                                "files": [
                                    {
                                        "sig_type": "efi",
                                        "file": "usr/lib/grub/grubx64.efi",
                                    }
                                ]
                            },
                        }
                    }
                ).encode()
            },
        )
        hello_path = directory / "hello_1.0-1_amd64.deb"
        ArtifactPlayground.write_deb_file(
            hello_path, data_files={PurePath("boot/vmlinuz"): b"vmlinuz"}
        )
        grub_path = directory / "grub-efi-amd64-bin_2.12-2_amd64.deb"
        ArtifactPlayground.write_deb_file(
            grub_path,
            data_files={PurePath("usr/lib/grub/grubx64.efi"): b"grub"},
        )

        # Set up a mock client.
        debusine_mock = self.mock_debusine()
        artifacts = {
            2: self._create_binary_package_response(
                2, "hello-signed-template", "1.0+1"
            ),
            3: self._create_binary_package_response(3, "hello", "1.0-1"),
            4: self._create_binary_package_response(
                4, "grub-efi-amd64-bin", "2.12-2"
            ),
        }
        artifact_files = {
            (2, "hello-signed-template_1.0+1_amd64.deb"): template_deb_path,
            (3, "hello_1.0-1_amd64.deb"): hello_path,
            (4, "grub-efi-amd64-bin_2.12-2_amd64.deb"): grub_path,
        }
        uploaded_artifact_ids = {
            ArtifactCategory.SIGNING_INPUT: 5,
            ArtifactCategory.WORK_REQUEST_DEBUG_LOGS: 6,
        }
        output_path = self.create_temporary_directory()

        def download_artifact(
            artifact_id: int,
            destination: Path,
            tarball: bool = False,
            workspace: str | None = None,  # noqa: ARG001, U100
        ) -> ArtifactResponse:
            assert not tarball
            for file_name in artifacts[artifact_id].files:
                target = destination / file_name
                target.parent.mkdir(parents=True, exist_ok=True)
                assert not target.exists()
                shutil.copy2(artifact_files[(artifact_id, file_name)], target)
            return artifacts[artifact_id]

        def upload_artifact(
            local_artifact: LocalArtifact[Any], **kwargs: Any
        ) -> RemoteArtifact:
            if local_artifact.category == ArtifactCategory.SIGNING_INPUT:
                for name, path in local_artifact.files.items():
                    assert not Path(name).is_absolute()
                    (output_path / name).parent.mkdir(
                        parents=True, exist_ok=True
                    )
                    shutil.copy(path, output_path / name)
            return create_remote_artifact(
                id=uploaded_artifact_ids[local_artifact.category],
                workspace="testing",
            )

        debusine_mock.download_artifact.side_effect = download_artifact
        debusine_mock.upload_artifact.side_effect = upload_artifact

        with (
            mock.patch.object(
                self.task, "_prepare_executor_instance", autospec=True
            ),
            mock.patch.object(
                self.task, "executor_instance", autospec=True
            ) as mock_executor_instance,
        ):
            mock_executor_instance.file_pull.side_effect = shutil.copy2
            mock_executor_instance.run.side_effect = _mock_run
            self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        self.assertEqual(debusine_mock.upload_artifact.call_count, 3)
        self.assertEqual(
            (output_path / "hello/boot/vmlinuz").read_bytes(), b"vmlinuz"
        )
        self.assertEqual(
            (
                output_path / "grub-efi-amd64-bin/usr/lib/grub/grubx64.efi"
            ).read_bytes(),
            b"grub",
        )
